"""
Interpolation and regridding utilities for skyborn.

This module provides various interpolation methods including:
- Nearest neighbor interpolation
- Bilinear interpolation
- Conservative interpolation
"""

from .regridding import (
    Grid,
    Regridder,
    NearestRegridder,
    BilinearRegridder,
    ConservativeRegridder,
    nearest_neighbor_indices,
    regrid_dataset,
)
from .interpolation import (
    interp_hybrid_to_pressure,
    interp_sigma_to_hybrid,
    interp_multidim,
)

__all__ = [
    "Grid",
    "Regridder",
    "NearestRegridder",
    "BilinearRegridder",
    "ConservativeRegridder",
    "nearest_neighbor_indices",
    "regrid_dataset",
    "interp_hybrid_to_pressure",
    "interp_sigma_to_hybrid",
    "interp_multidim",
]



"""
Interpolation functions for hybrid-sigma and multidimensional data.

References
----------
geocat-comp interpolation:
    https://github.com/NCAR/geocat-comp/blob/main/geocat/comp/interpolation.py
"""

import typing
import warnings
import metpy.interpolate
import numpy as np
import xarray as xr

supported_types = typing.Union[xr.DataArray, np.ndarray]

__pres_lev_mandatory__ = np.array(
    [
        1000,
        925,
        850,
        700,
        500,
        400,
        300,
        250,
        200,
        150,
        100,
        70,
        50,
        30,
        20,
        10,
        7,
        5,
        3,
        2,
        1,
    ]
).astype(
    np.float32
)  # Mandatory pressure levels (mb)
__pres_lev_mandatory__ = __pres_lev_mandatory__ * 100.0  # Convert mb to Pa


def _func_interpolate(method="linear"):
    """Define interpolation function."""

    if method == "linear":
        func_interpolate = metpy.interpolate.interpolate_1d
    elif method == "log":
        func_interpolate = metpy.interpolate.log_interpolate_1d
    else:
        raise ValueError(
            f"Unknown interpolation method: {method}. "
            f'Supported methods are: "log" and "linear".'
        )

    return func_interpolate


def _pressure_from_hybrid(psfc, hya, hyb, p0=100000.0):
    """Calculate pressure at the hybrid levels."""

    # p(k) = hya(k) * p0 + hyb(k) * psfc

    # This will be in Pa
    return hya * p0 + hyb * psfc


def _pre_interp_multidim(
    data_in: xr.DataArray,
    cyclic: bool,
    missing_val,
):
    """Helper Function: Handling missing data functionality and adding cyclic
    point if required.

    Parameters
    ----------
    data_in : :class:`xarray.DataArray`
        The data on which to operate

    cyclic : :class:`bool`
        Determines if cyclic point should be added or not.
        If true then add point, else do nothing.

    missing_val : int, float, optional
        Provides an alternative to NaN

    Returns
    -------
    data_in : :class:`xarray.DataArray`
       The data input with cyclic points added (if cyclic is true)
       and missing_val values replaced with np.nan

    Notes
    -------
    """
    # replace missing_val with np.nan
    if missing_val is not None:
        data_in = xr.DataArray(
            np.where(data_in.values == missing_val, np.nan, data_in.values),
            dims=data_in.dims,
            coords=data_in.coords,
        )

    # add cyclic points and create new data array
    if cyclic:
        padded_data = np.pad(data_in.values, ((0, 0), (1, 1)), mode="wrap")
        padded_longitudes = np.pad(
            data_in.coords[data_in.dims[-1]], (1, 1), mode="wrap"
        )
        padded_longitudes[0] -= 360
        padded_longitudes[-1] += 360

        data_in = xr.DataArray(
            padded_data,
            coords={
                data_in.dims[-2]: data_in.coords[data_in.dims[-2]].values,
                data_in.dims[-1]: padded_longitudes,
            },
            dims=data_in.dims,
        )

    return data_in


def _post_interp_multidim(data_in, missing_val):
    """Helper Function: Handling missing data functionality.

    Parameters
    ----------
    data_in : :class:`xarray.DataArray`
        The data on which to operate

    missing_val : int, float, optional
         Provides an alternative to NaN

    Returns
    -------
    data_in : :class:`xarray.DataArray`
       The data input with np.nan values replaced with missing_val
    """
    if missing_val is not None:
        data_in = xr.DataArray(
            np.where(np.isnan(data_in.values), missing_val, data_in.values),
            dims=data_in.dims,
            coords=data_in.coords,
        )

    return data_in


def _sigma_from_hybrid(psfc, hya, hyb, p0=100000.0):
    """Calculate sigma at the hybrid levels."""

    # sig(k) = hya(k) * p0 / psfc + hyb(k)

    # This will be in Pa
    return hya * p0 / psfc + hyb


def _vertical_remap(func_interpolate, new_levels, xcoords, data, interp_axis=0):
    """Execute the defined interpolation function on data."""

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", r"Interpolation point out of data bounds encountered"
        )
        return func_interpolate(new_levels, xcoords, data, axis=interp_axis)


def _temp_extrapolate(data, lev_dim, lev, p_sfc, ps, phi_sfc):
    r"""This helper function extrapolates temperature below ground using the
    ECMWF formulation described in `Vertical Interpolation and Truncation of
    Model-Coordinate Data <https://dx.doi.org/10.5065/D6HX19NH>`__ by Trenberth,
    Berry, & Buja [NCAR/TN-396, 1993]. Specifically equation 16 is used:

    .. math::
        T = T_* \left( 1 + \alpha ln \frac{p}{p_s} + \frac{1}{2}\left( \alpha ln \frac{p}{p_s} \right)^2 + \frac{1}{6} \left( \alpha ln \frac{p}{p_s} \right)^3 \right)

    Parameters
    ----------
    data: :class:`xarray.DataArray`
        The temperature at the lowest level of the model.

    lev_dim: str
        The name of the vertical dimension.

    lev: int
        The pressure levels of interest. Must be in the same units as ``ps`` and ``p_sfc``

    p_sfc: :class:`xarray.DataArray`
        The pressure at the lowest level of the model. Must be in the same units as ``lev`` and ``ps``

    ps: :class:`xarray.DataArray`
        An array of surface pressures. Must be in the same units as ``lev`` and ``p_sfc``

    phi_sfc: :class:`xarray.DataArray`
        The geopotential at the lowest level of the model.

    Returns
    -------
    result: :class:`xarray.DataArray`
        The extrapolated temperatures at the provided pressure levels.
    """
    R_d = 287.04  # dry air gas constant
    g_inv = 1 / 9.80616  # inverse of gravity
    alpha = 0.0065 * R_d * g_inv

    tstar = data.isel({lev_dim: -1}, drop=True) * (1 + alpha * (ps / p_sfc - 1))
    hgt = phi_sfc * g_inv
    t0 = tstar + 0.0065 * hgt
    tplat = xr.apply_ufunc(np.minimum, 298, t0, dask="parallelized")

    tprime0 = xr.where(
        (2000 <= hgt) & (hgt <= 2500),
        0.002 * ((2500 - hgt) * t0 + ((hgt - 2000) * tplat)),
        np.nan,
    )
    tprime0 = xr.where(2500 < hgt, tplat, tprime0)

    alnp = xr.where(
        hgt < 2000,
        alpha * np.log(lev / ps),
        R_d * (tprime0 - tstar) / phi_sfc * np.log(lev / ps),
    )
    alnp = xr.where(tprime0 < tstar, 0, alnp)

    return tstar * (1 + alnp + (0.5 * (alnp**2)) + (1 / 6 * (alnp**3)))


def _geo_height_extrapolate(t_bot, lev, p_sfc, ps, phi_sfc):
    r"""This helper function extrapolates geopotential height below ground using
    the ECMWF formulation described in `Vertical Interpolation and Truncation
    of Model-Coordinate Data <https://dx.doi.org/10.5065/D6HX19NH>`__ by
    Trenberth, Berry, & Buja [NCAR/TN-396, 1993]. Specifically equation 15 is
    used:

    .. math::
        \Phi = \Phi_s - R_d T_* ln \frac{p}{p_s} \left[ 1 + \frac{1}{2}\alpha ln\frac{p}{p_s} + \frac{1}{6} \left( \alpha ln \frac{p}{p_s} \right)^2 \right]

    Parameters
    ----------
    t_bot: :class:`xarray.DataArray`
        Temperature at the lowest (bottom) level of the model.

    lev: int
        The pressure level of interest. Must be in the same units as ``ps`` and ``p_sfc``

    p_sfc: :class:`xarray.DataArray`
        The pressure at the lowest level of the model. Must be in the same units as ``lev`` and ``ps``

    ps : :class:`xarray.DataArray`
        An array of surface pressures. Must be in the same units as ``lev`` and ``p_sfc``

    phi_sfc:
        The geopotential at the lowest level of the model.

    Returns
    -------
    result: :class:`xarray.DataArray`
        The extrapolated geopotential height in geopotential meters at the provided pressure levels.
    """
    R_d = 287.04  # dry air gas constant
    g_inv = 1 / 9.80616  # inverse of gravity
    alpha = 0.0065 * R_d * g_inv

    tstar = t_bot * (1 + alpha * (ps / p_sfc - 1))
    hgt = phi_sfc * g_inv
    t0 = tstar + 0.0065 * hgt

    alph = xr.where(
        (tstar <= 290.5) & (t0 > 290.5), R_d / phi_sfc * (290.5 - tstar), alpha
    )

    alph = xr.where((tstar > 290.5) & (t0 > 290.5), 0, alph)
    tstar = xr.where((tstar > 290.5) & (t0 > 290.5), 0.5 * (290.5 + tstar), tstar)

    tstar = xr.where((tstar < 255), 0.5 * (tstar + 255), tstar)

    alnp = alph * np.log(lev / ps)
    return hgt - R_d * tstar * g_inv * np.log(lev / ps) * (
        1 + 0.5 * alnp + 1 / 6 * alnp**2
    )


def _vertical_remap_extrap(
    new_levels, lev_dim, data, output, pressure, ps, variable, t_bot, phi_sfc
):
    """A helper function to call the appropriate extrapolation function based
    on the user's inputs.

    Parameters
    ----------
    new_levels: array-like
        The desired pressure levels for extrapolation in Pascals.

    lev_dim: str
        The name of the vertical dimension.

    data: :class:`xarray.DataArray`
        The data to extrapolate

    output: :class:`xarray.DataArray`
        An array to hold the output data

    pressure: :class:`xarray.DataArray`
        The pressure at the lowest level of the model. Must be in the same units as ``lev`` and ``ps``

    ps : :class:`xarray.DataArray`
        An array of surface pressures. Must be in the same units as ``lev`` and ``p_sfc``

    variable : str, optional
        String representing what variable is extrapolated below surface level.
        Temperature extrapolation = "temperature". Geopotential height
        extrapolation = "geopotential". All other variables = "other". If
        "other", the value of ``data`` at the lowest model level will be used
        as the below ground fill value. Required if extrapolate is True.

    t_bot: :class:`xarray.DataArray`
        Temperature at the lowest (bottom) level of the model.

    phi_sfc:
        The geopotential at the lowest level of the model.

    Returns
    -------
    output: :class:`xarray.DataArray`
        A DataArray containing the data after extrapolation.
    """

    sfc_index = pressure[lev_dim].argmax(dim=lev_dim)  # index of the model surface
    p_sfc = pressure.isel(
        {lev_dim: sfc_index}, drop=True
    )  # extract pressure at lowest level

    if variable == "temperature":
        output = output.where(
            output.plev <= p_sfc,
            _temp_extrapolate(data, lev_dim, output.plev, p_sfc, ps, phi_sfc),
        )
    elif variable == "geopotential":
        output = output.where(
            output.plev <= p_sfc,
            _geo_height_extrapolate(t_bot, output.plev, p_sfc, ps, phi_sfc),
        )
    else:
        output = output.where(
            output.plev <= p_sfc, data.isel({lev_dim: sfc_index}, drop=True)
        )

    return output


def interp_hybrid_to_pressure(
    data: xr.DataArray,
    ps: xr.DataArray,
    hyam: xr.DataArray,
    hybm: xr.DataArray,
    p0: float = 100000.0,
    new_levels: np.ndarray = __pres_lev_mandatory__,
    lev_dim: str = None,
    method: str = "linear",
    extrapolate: bool = False,
    variable: str = None,
    t_bot: xr.DataArray = None,
    phi_sfc: xr.DataArray = None,
) -> xr.DataArray:
    """Interpolate and extrapolate data from hybrid-sigma levels to isobaric
    levels. Keeps attributes (i.e. metadata) of the input data in the output as
    default.

    Notes
    -----
    Atmosphere hybrid-sigma pressure coordinates are commonly defined in two different
    ways as described below and in CF Conventions. This particular function expects the
    first formulation. However, with some minor adjustments on the user side it can
    support datasets leveraging the second formulation as well.  In this case, you can
    set the input parameters p0=1 and hyam=ap to adapt the function to meet your needs.

    Formulation 1: p(n,k,j,i) = a(k)*p0 + b(k)*ps(n,j,i)
    Formulation 2: p(n,k,j,i) = ap(k) + b(k)*ps(n,j,i)

    ACKNOWLEDGEMENT: We'd like to thank to `Brian Medeiros <https://github.com/brianpm>`__,
    `Matthew Long <https://github.com/matt-long>`__, and `Deepak Cherian <https://github.com/dcherian>`__
    at NSF NCAR for their great contributions since the code implemented here is mostly
    based on their work.

    Parameters
    ----------
    data : :class:`xarray.DataArray`
        Multidimensional data array of hybrid-sigma levels and has a ``lev_dim`` coordinate.

    ps : :class:`xarray.DataArray`
        A multi-dimensional array of surface pressures (Pa), same time/space shape as data.

    hyam, hybm : :class:`xarray.DataArray`
        One-dimensional arrays containing the hybrid A and B coefficients. Must have the same
        dimension size as the ``lev_dim`` dimension of data.

    p0 : float, optional
        Scalar numeric value equal to surface reference pressure (Pa). Defaults to 100000 Pa.

    new_levels : ndarray, optional
        A one-dimensional array of output pressure levels (Pa). If not given, the mandatory
        list of 21 pressure levels is used.

    lev_dim : str, optional
        String that is the name of level dimension in data. Defaults to "lev".

    method : str, optional
        String that is the interpolation method; can be either "linear" or "log". Defaults to "linear".

    extrapolate : bool, optional
        If True, below ground extrapolation for ``variable`` will be done using
        an `ECMWF formulation <https://dx.doi.org/10.5065/D6HX19NH>`__. Defaults
        to False.

    variable : str, optional
        String representing what variable is extrapolated below surface level.
        Temperature extrapolation = "temperature". Geopotential height
        extrapolation = "geopotential". All other variables = "other". If
        "other", the value of ``data`` at the lowest model level will be used
        as the below ground fill value. Required if extrapolate is True.

    t_bot : :class:`xarray.DataArray`, optional
        Temperature in Kelvin at the lowest layer of the model. Not necessarily
        the same as surface temperature. Required if ``extrapolate`` is True
        and ``variable`` is not ``'other'``

    phi_sfc: :class:`xarray.DataArray`, optional
        Geopotential in J/kg at the lowest layer of the model. Not necessarily
        the same as surface geopotential. Required if ``extrapolate`` is True
        and ``variable`` is not ``'other'``.

    Returns
    -------
    output : :class:`xarray.DataArray`
        Interpolated data with isobaric levels

    See Also
    --------
    Related NCL Functions:
    `vinth2p <https://www.ncl.ucar.edu/Document/Functions/Built-in/vinth2p.shtml>`__,
    `vinth2p_ecmwf <https://www.ncl.ucar.edu/Document/Functions/Built-in/vinth2p_ecmwf.shtml>`__
    """

    # Check inputs
    if extrapolate and (variable is None):
        raise ValueError("If `extrapolate` is True, `variable` must be provided.")

    if variable in ["geopotential", "temperature"] and (
        t_bot is None or phi_sfc is None
    ):
        raise ValueError(
            "If `variable` is 'geopotential' or 'temperature', both `t_bot` and `phi_sfc` must be provided"
        )

    if variable not in ["geopotential", "temperature", "other", None]:
        raise ValueError(
            "The value of `variable` is "
            + variable
            + ", but the accepted values are 'temperature', 'geopotential', 'other', or None."
        )

    # Determine the level dimension and then the interpolation axis
    if lev_dim is None:
        try:
            lev_dim = data.cf["vertical"].name
        except Exception:
            raise ValueError(
                "Unable to determine vertical dimension name. Please specify the name via `lev_dim` argument."
            )

    try:
        func_interpolate = _func_interpolate(method)
    except ValueError as vexc:
        raise ValueError(vexc.args[0])

    interp_axis = data.dims.index(lev_dim)

    # Calculate pressure levels at the hybrid levels
    pressure = _pressure_from_hybrid(ps, hyam, hybm, p0)  # Pa

    # Make pressure shape same as data shape
    pressure = pressure.transpose(*data.dims)

    ###############################################################################
    # Workaround
    #
    # For the issue with metpy's xarray interface:
    #
    # `metpy.interpolate.interpolate_1d` had "no implementation found for
    # 'numpy.apply_along_axis'" issue for cases where the input is
    # xarray.Dataarray and has more than 3 dimensions (e.g. 4th dim of `time`).

    # Use dask.array.core.map_blocks instead of xarray.apply_ufunc and
    # auto-chunk input arrays to ensure using only Numpy interface of
    # `metpy.interpolate.interpolate_1d`.

    # # Apply vertical interpolation
    # # Apply Dask parallelization with xarray.apply_ufunc
    # output = xr.apply_ufunc(
    #     _vertical_remap,
    #     data,
    #     pressure,
    #     exclude_dims=set((lev_dim,)),  # Set dimensions allowed to change size
    #     input_core_dims=[[lev_dim], [lev_dim]],  # Set core dimensions
    #     output_core_dims=[["plev"]],  # Specify output dimensions
    #     vectorize=True,  # loop over non-core dims
    #     dask="parallelized",  # Dask parallelization
    #     output_dtypes=[data.dtype],
    #     dask_gufunc_kwargs={"output_sizes": {
    #         "plev": len(new_levels)
    #     }},
    # )

    # If an unchunked Xarray input is given, chunk it just with its dims
    if data.chunks is None:
        data_chunk = dict([(k, v) for (k, v) in zip(list(data.dims), list(data.shape))])
        data = data.chunk(data_chunk)

    # Chunk pressure equal to data's chunks
    pressure = pressure.chunk(data.chunksizes)

    # Output data structure elements
    out_chunks = list(data.chunks)
    out_chunks[interp_axis] = (new_levels.size,)
    out_chunks = tuple(out_chunks)
    # ''' end of boilerplate

    from dask.array.core import map_blocks

    output = map_blocks(
        _vertical_remap,
        func_interpolate,
        new_levels,
        pressure.data,
        data.data,
        interp_axis,
        chunks=out_chunks,
        dtype=data.dtype,
        drop_axis=[interp_axis],
        new_axis=[interp_axis],
    )

    # End of Workaround
    ###############################################################################

    output = xr.DataArray(output, name=data.name, attrs=data.attrs)

    # Set output dims and coords
    dims = [data.dims[i] if i != interp_axis else "plev" for i in range(data.ndim)]

    # Rename output dims. This is only needed with above workaround block
    dims_dict = {output.dims[i]: dims[i] for i in range(len(output.dims))}
    output = output.rename(dims_dict)

    coords = {}
    for k, v in data.coords.items():
        if k != lev_dim:
            coords.update({k: v})
        else:
            coords.update({"plev": new_levels})

    output = output.transpose(*dims).assign_coords(coords)

    if extrapolate:
        output = _vertical_remap_extrap(
            new_levels, lev_dim, data, output, pressure, ps, variable, t_bot, phi_sfc
        )

    return output


def interp_sigma_to_hybrid(
    data: xr.DataArray,
    sig_coords: xr.DataArray,
    ps: xr.DataArray,
    hyam: xr.DataArray,
    hybm: xr.DataArray,
    p0: float = 100000.0,
    lev_dim: str = None,
    method: str = "linear",
) -> xr.DataArray:
    """Interpolate data from sigma to hybrid coordinates.  Keeps the attributes
    (i.e. meta information) of the input data in the output as default.

    Parameters
    ----------
    data : :class:`xarray.DataArray`
        Multidimensional data array, which holds sigma levels and has a ``lev_dim`` coordinate.

    sig_coords : :class:`xarray.DataArray`
        A one-dimensional array of sigma coordinates of ``lev_dim`` of ``data``.

    ps : :class:`xarray.DataArray`
        A multi-dimensional array of surface pressures (Pa), same time/space shape as data.

    hyam, hybm : :class:`xarray.DataArray`
        One-dimensional arrays containing the hybrid A and B coefficients. Must have the same
        dimension as the output hybrid levels.

    p0 : float, optional
        Scalar numeric value equal to surface reference pressure (Pa). Defaults to 100000 Pa.

    lev_dim : str, optional
        String that is the name of level dimension in data. Defaults to "lev".

    method : str, optional
        String that is the interpolation method; can be either "linear" or "log". Defaults to "linear".

    Returns
    -------
    output : :class:`xarray.DataArray`
        Interpolated data with hybrid levels

    See Also
    --------
    Related NCL Function:
    `sigma2hybrid <https://www.ncl.ucar.edu/Document/Functions/Built-in/sigma2hybrid.shtml>`__
    """

    # Determine the level dimension and then the interpolation axis
    if lev_dim is None:
        try:
            lev_dim = data.cf["vertical"].name
        except Exception:
            raise ValueError(
                "Unable to determine vertical dimension name. Please specify the name via `lev_dim` argument.'"
            )

    try:
        func_interpolate = _func_interpolate(method)
    except ValueError as vexc:
        raise ValueError(vexc.args[0])

    # Calculate sigma levels at the hybrid levels
    sigma = _sigma_from_hybrid(ps, hyam, hybm, p0)  # Pa

    non_lev_dims = list(data.dims)
    if data.ndim > 1:
        non_lev_dims.remove(lev_dim)
        data_stacked = data.stack(combined=non_lev_dims).transpose()
        sigma_stacked = sigma.stack(combined=non_lev_dims).transpose()

        h_coords = sigma_stacked[0, :].copy()

        output = data_stacked[:, : len(hyam)].copy()

        for idx, (d, s) in enumerate(zip(data_stacked, sigma_stacked)):
            output[idx, :] = xr.DataArray(
                _vertical_remap(func_interpolate, s.data, sig_coords.data, d.data),
                dims=[lev_dim],
            )

        # Make output shape same as data shape
        output = output.unstack().transpose(*data.dims)
    else:
        h_coords = sigma

        output = data[: len(hyam)].copy()
        output[: len(hyam)] = xr.DataArray(
            _vertical_remap(func_interpolate, sigma.data, sig_coords.data, data.data),
            dims=[lev_dim],
        )

    # Set output dims and coords
    output = output.rename({lev_dim: "hlev"})
    output = output.assign_coords({"hlev": h_coords.data})

    return output


def interp_multidim(
    data_in: supported_types,
    lat_out: np.ndarray,
    lon_out: np.ndarray,
    lat_in: np.ndarray = None,
    lon_in: np.ndarray = None,
    cyclic: bool = False,
    missing_val: np.number = None,
    method: str = "linear",
    fill_value: typing.Union[str, np.number] = np.nan,
) -> supported_types:
    """Multidimensional interpolation of variables. Uses ``xarray.interp`` to
    perform interpolation. Will not perform extrapolation by default, returns
    missing values if any surrounding points contain missing values.

    .. warning::
        The output data type may be promoted to that of the coordinate data.

    Parameters
    ----------
    data_in : :class:`xarray.DataArray`, ndarray
        Data array with data to be interpolated and associated coords. If
        it is a np array, then ``lat_in`` and ``lon_in`` must be provided. Length must
        be coordinated with given coordinates.

    lat_out: ndarray
        List of latitude coordinates to be interpolated to.

    lon_out: ndarray
        List of longitude coordinates to be interpolated to.

    lat_in: ndarray
        List of latitude coordinates corresponding to ``data_in``. Must be
        given if ``data_in`` is not an xarray.

    lon_in: ndarray
        List of longitude coordinates corresponding to ``data_in``. Must be
        given if ``data_in`` is not an xarray.

    cyclic: bool, optional
        Set as true if lon values are cyclical but do not fully wrap around
        the globe
        (0, 1.5, 3, ..., 354, 355.5) Default is false

    missing_val : :class:`np.number`, optional
        Provide a number to represent missing data. Alternative to using ``np.nan``

    method: str, optional
        Provide specific method of interpolation. Default is "linear"
        “linear” or “nearest” for multidimensional array

    fill_value: str, optional
        Set as 'extrapolate' to allow extrapolation of data. Default is
        no extrapolation.

    Returns
    -------
    data_out : ndarray, :class:`xarray.DataArray`
       Returns the same type of object as input ``data_in``. However, the type of
       the data in the array may be promoted to that of the coordinates. Shape
       will be the same as input array except for last two dimensions which will
       be equal to the coordinates given in ``data_out``.

    Examples
    --------
    >>> import xarray as xr
    >>> import numpy as np
    >>> import geocat.comp
    >>> data = np.asarray([[1, 2, 3, 4, 5, 99],
    ...                   [2, 4, 6, 8, 10, 12]])
    >>> lat_in = [0, 1]
    >>> lon_in = [0, 50, 100, 250, 300, 350]
    >>> data_in = xr.DataArray(data,
    ...                        dims=['lat', 'lon'],
    ...                        coords={'lat':lat_in,
    ...                                'lon': lon_in})
    >>> data_out = xr.DataArray(dims=['lat', 'lon'],
    ...                         coords={'lat': [0, 1],
    ...                                 'lon': [0, 50, 360]})
    >>> do = interp_multidim(data_in,
    ...                      [0, 1],
    ...                      [0, 50, 360],
    ...                      cyclic=True,
    ...                      missing_val=99)
    >>> print(do)
    <xarray.DataArray (lat: 2, lon: 3)>
    array([[ 1.,  2., 99.],
       [ 2.,  4., 99.]])
    Coordinates:
      * lat      (lat) int64 0 1
      * lon      (lon) int64 0 50 360

    See Also
    --------
    Related External Functions:
    `xarray.DataArray.interp <https://docs.xarray.dev/en/stable/generated/xarray.DataArray.interp.html>`__,
    `cartopy.util.add_cyclic_point <https://scitools.org.uk/cartopy/docs/latest/reference/generated/cartopy.util.add_cyclic_point.html>`__

    Related NCL Function:
    `NCL linint2 <https://www.ncl.ucar.edu/Document/Functions/Built-in/linint2.shtml>`__
    """
    # check for xarray/numpy
    if not isinstance(data_in, xr.DataArray):
        if lat_in is None or lon_in is None:
            raise ValueError(
                "Argument lat_in and lon_in must be provided if data_in is not an xarray"
            )
        data_in = xr.DataArray(
            data_in, dims=["lat", "lon"], coords={"lat": lat_in, "lon": lon_in}
        )

    output_coords = {
        data_in.dims[-1]: lon_out,
        data_in.dims[-2]: lat_out,
    }

    data_in_modified = _pre_interp_multidim(data_in, cyclic, missing_val)
    data_out = data_in_modified.interp(
        output_coords, method=method, kwargs={"fill_value": fill_value}
    )
    data_out_modified = _post_interp_multidim(data_out, missing_val=missing_val)

    return data_out_modified



"""
This scripts contains functions that performs nearest, bilinear, and conservative interpolation
on xarray.Datasets. The original version of this script is available at WeatherBench2.

Qianye Su
suqianye2000@gmail.com

Reference
 - WeatherBench2 regridding:
     https://github.com/google-research/weatherbench2/blob/main/weatherbench2/regridding.py
"""

from __future__ import annotations

import xarray
import numpy as np
from sklearn import neighbors

import functools
import dataclasses
from typing import Union, Tuple, Optional

Array = Union[np.ndarray]


def _detect_coordinate_names(dataset: xarray.Dataset) -> Tuple[str, str]:
    """
    Detect latitude and longitude coordinate names in the dataset.

    Args:
        dataset: xarray Dataset

    Returns:
        Tuple of (longitude_name, latitude_name)

    Raises:
        ValueError: If coordinate names cannot be detected
    """
    # Common variations of coordinate names
    lon_names = ["longitude", "lon", "long", "x"]
    lat_names = ["latitude", "lat", "y"]

    # Find longitude coordinate
    lon_coord = None
    for name in lon_names:
        if name in dataset.dims:
            lon_coord = name
            break

    # Find latitude coordinate
    lat_coord = None
    for name in lat_names:
        if name in dataset.dims:
            lat_coord = name
            break

    if lon_coord is None or lat_coord is None:
        available_dims = list(dataset.dims.keys())
        raise ValueError(
            f"Could not detect longitude/latitude coordinates. "
            f"Available dimensions: {available_dims}. "
            f"Expected one of: lon={lon_names}, lat={lat_names}"
        )

    return lon_coord, lat_coord


@dataclasses.dataclass(frozen=True)
class Grid:
    """Representation of a rectilinear grid."""

    lon: np.ndarray
    lat: np.ndarray

    @classmethod
    def from_degrees(cls, lon: np.ndarray, lat: np.ndarray) -> Grid:
        return cls(np.deg2rad(lon), np.deg2rad(lat))

    @classmethod
    def from_dataset(cls, dataset: xarray.Dataset) -> Grid:
        """Create a Grid from an xarray Dataset by auto-detecting coordinates."""
        lon_name, lat_name = _detect_coordinate_names(dataset)
        lon_values = dataset[lon_name].values
        lat_values = dataset[lat_name].values
        return cls.from_degrees(lon_values, lat_values)

    @property
    def shape(self) -> tuple[int, int]:
        return (len(self.lon), len(self.lat))

    def _to_tuple(self) -> tuple[tuple[float, ...], tuple[float, ...]]:
        return tuple(self.lon.tolist()), tuple(self.lat.tolist())

    def __eq__(self, other):  # needed for hashability
        return isinstance(other, Grid) and self._to_tuple() == other._to_tuple()

    def __hash__(self):
        return hash(self._to_tuple())


@dataclasses.dataclass(frozen=True)
class Regridder:
    """Base class for regridding."""

    source: Grid
    target: Grid

    def regrid_array(self, field: Array) -> np.ndarray:
        """Regrid an array with dimensions (..., lon, lat) from source to target."""
        raise NotImplementedError

    def regrid_dataset(
        self,
        dataset: xarray.Dataset,
        lon_dim: Optional[str] = None,
        lat_dim: Optional[str] = None,
    ) -> xarray.Dataset:
        """
        Regrid an xarray.Dataset from source to target.

        Args:
            dataset: Input xarray Dataset
            lon_dim: Name of longitude dimension (auto-detected if None)
            lat_dim: Name of latitude dimension (auto-detected if None)

        Returns:
            Regridded xarray Dataset
        """
        # Auto-detect coordinate names if not provided
        if lon_dim is None or lat_dim is None:
            detected_lon, detected_lat = _detect_coordinate_names(dataset)
            lon_dim = lon_dim or detected_lon
            lat_dim = lat_dim or detected_lat

        # Ensure latitude is in ascending order
        if not (dataset[lat_dim].diff(lat_dim) > 0).all():
            dataset = dataset.isel({lat_dim: slice(None, None, -1)})  # Reverse
        assert (dataset[lat_dim].diff(lat_dim) > 0).all()

        # Apply regridding
        dataset = xarray.apply_ufunc(
            self.regrid_array,
            dataset,
            input_core_dims=[[lon_dim, lat_dim]],
            output_core_dims=[[lon_dim, lat_dim]],
            exclude_dims={lon_dim, lat_dim},
            vectorize=True,
            dask="allowed",  # Allow Dask arrays to be passed to the function
            output_dtypes=[dataset[list(dataset.data_vars)[0]].dtype],
        )

        # Update coordinates to target grid
        target_lon_deg = np.rad2deg(self.target.lon)
        target_lat_deg = np.rad2deg(self.target.lat)
        dataset = dataset.assign_coords(
            {lon_dim: target_lon_deg, lat_dim: target_lat_deg}
        )

        return dataset


def nearest_neighbor_indices(source_grid: Grid, target_grid: Grid) -> np.ndarray:
    """Returns Haversine nearest neighbor indices from source_grid to target_grid."""
    # Construct a BallTree to find nearest neighbors on the sphere
    source_mesh = np.meshgrid(source_grid.lat, source_grid.lon, indexing="ij")
    target_mesh = np.meshgrid(target_grid.lat, target_grid.lon, indexing="ij")
    index_coords = np.stack([x.ravel() for x in source_mesh], axis=-1)
    query_coords = np.stack([x.ravel() for x in target_mesh], axis=-1)
    tree = neighbors.BallTree(index_coords, metric="haversine")
    indices = tree.query(query_coords, return_distance=False).squeeze(axis=-1)
    return indices


class NearestRegridder(Regridder):
    """Regrid with nearest neighbor interpolation."""

    @functools.cached_property
    def indices(self):
        """The interpolation indices associated with source_grid."""
        return nearest_neighbor_indices(self.source, self.target)

    def _nearest_neighbor_2d(self, array: Array) -> np.ndarray:
        """2D nearest neighbor interpolation using BallTree."""
        if array.shape != self.source.shape:
            raise ValueError(
                f"Expected array.shape={array.shape} to match source.shape={self.source.shape}"
            )
        array_flat = array.ravel()
        interpolated = array_flat[self.indices]
        return interpolated.reshape(self.target.shape)

    def regrid_array(self, field: Array) -> np.ndarray:
        interp = np.vectorize(self._nearest_neighbor_2d, signature="(a,b)->(c,d)")
        return interp(field)


class BilinearRegridder(Regridder):
    """Regrid with bilinear interpolation."""

    def regrid_array(self, field: Array) -> np.ndarray:
        lat_source = self.source.lat
        lat_target = self.target.lat
        lon_source = self.source.lon
        lon_target = self.target.lon

        # Interpolate over latitude
        lat_interp = np.array(
            [
                np.interp(lat_target, lat_source, f)
                for f in field.transpose(1, 0, *range(2, field.ndim))
            ]
        ).transpose(1, 0, *range(2, field.ndim))

        # Interpolate over longitude
        lon_interp = np.array(
            [
                np.interp(lon_target, lon_source, f)
                for f in lat_interp.transpose(0, *range(2, field.ndim), 1)
            ]
        ).transpose(0, *range(2, field.ndim), 1)

        return lon_interp


def _assert_increasing(x: np.ndarray) -> None:
    if not (np.diff(x) > 0).all():
        raise ValueError(f"Array is not increasing: {x}")


def _latitude_cell_bounds(x: Array) -> np.ndarray:
    pi_over_2 = np.array([np.pi / 2], dtype=x.dtype)
    return np.concatenate((-pi_over_2, (x[:-1] + x[1:]) / 2, pi_over_2))


def _latitude_overlap(
    source_points: Array,
    target_points: Array,
) -> np.ndarray:
    """Calculate the area overlap as a function of latitude."""
    source_bounds = _latitude_cell_bounds(source_points)
    target_bounds = _latitude_cell_bounds(target_points)
    upper = np.minimum(target_bounds[1:, np.newaxis], source_bounds[np.newaxis, 1:])
    lower = np.maximum(target_bounds[:-1, np.newaxis], source_bounds[np.newaxis, :-1])
    # Normalized cell area: integral from lower to upper of cos(latitude)
    overlap = (upper > lower) * (np.sin(upper) - np.sin(lower))
    return overlap


def _conservative_latitude_weights(
    source_points: Array, target_points: Array
) -> np.ndarray:
    """Create a weight matrix for conservative regridding along latitude.

    Args:
        source_points: 1D latitude coordinates in radians for centers of source cells.
        target_points: 1D latitude coordinates in radians for centers of target cells.

    Returns:
        NumPy array with shape (target_size, source_size). Rows sum to 1.
    """
    _assert_increasing(source_points)
    _assert_increasing(target_points)
    weights = _latitude_overlap(source_points, target_points)
    weights /= np.sum(weights, axis=1, keepdims=True)
    return weights


def _align_phase_with(x, target, period):
    """Align the phase of a periodic number to match another."""
    shift_down = x > target + period / 2
    shift_up = x < target - period / 2
    return x + period * shift_up - period * shift_down


def _periodic_upper_bounds(x, period):
    x_plus = _align_phase_with(np.roll(x, -1), x, period)
    return (x + x_plus) / 2


def _periodic_lower_bounds(x, period):
    x_minus = _align_phase_with(np.roll(x, +1), x, period)
    return (x_minus + x) / 2


def _periodic_overlap(x0, x1, y0, y1, period):
    """Calculate the overlap between two intervals considering periodicity."""
    y0 = _align_phase_with(y0, x0, period)
    y1 = _align_phase_with(y1, x0, period)
    upper = np.minimum(x1, y1)
    lower = np.maximum(x0, y0)
    return np.maximum(upper - lower, 0)


def _longitude_overlap(
    first_points: Array,
    second_points: Array,
    period: float = 2 * np.pi,
) -> np.ndarray:
    """Calculate the area overlap as a function of longitude."""
    first_points = first_points % period
    first_upper = _periodic_upper_bounds(first_points, period)
    first_lower = _periodic_lower_bounds(first_points, period)

    second_points = second_points % period
    second_upper = _periodic_upper_bounds(second_points, period)
    second_lower = _periodic_lower_bounds(second_points, period)

    x0 = first_lower[:, np.newaxis]
    x1 = first_upper[:, np.newaxis]
    y0 = second_lower[np.newaxis, :]
    y1 = second_upper[np.newaxis, :]

    overlap_func = np.vectorize(_periodic_overlap, excluded=["period"])
    overlap = overlap_func(x0, x1, y0, y1, period=period)
    return overlap


def _conservative_longitude_weights(
    source_points: np.ndarray, target_points: np.ndarray
) -> np.ndarray:
    """Create a weight matrix for conservative regridding along longitude.

    Args:
        source_points: 1D longitude coordinates in radians for centers of source cells.
        target_points: 1D longitude coordinates in radians for centers of target cells.

    Returns:
        NumPy array with shape (target_size, source_size). Rows sum to 1.
    """
    _assert_increasing(source_points)
    _assert_increasing(target_points)
    weights = _longitude_overlap(target_points, source_points)
    weights /= np.sum(weights, axis=1, keepdims=True)
    return weights


class ConservativeRegridder(Regridder):
    """Regrid with linear conservative regridding."""

    def _mean(self, field: Array) -> np.ndarray:
        """Computes cell-averages of field on the target grid."""
        lon_weights = _conservative_longitude_weights(self.source.lon, self.target.lon)
        lat_weights = _conservative_latitude_weights(self.source.lat, self.target.lat)
        result = np.einsum(
            "ac,bd,...cd->...ab",
            lon_weights,
            lat_weights,
            field,
            optimize=True,
        )
        return result

    def _nanmean(self, field: Array) -> np.ndarray:
        """Compute cell-averages skipping NaNs like np.nanmean."""
        nulls = np.isnan(field)
        total = self._mean(np.where(nulls, 0, field))
        count = self._mean(~nulls)
        with np.errstate(divide="ignore", invalid="ignore"):
            result = np.true_divide(total, count)
            result[count == 0] = np.nan  # Set divisions by zero to NaN
        return result

    def regrid_array(self, field: Array) -> np.ndarray:
        return self._nanmean(field)


# Convenience function for easy regridding
def regrid_dataset(
    dataset: xarray.Dataset,
    target_grid: Grid,
    method: str = "bilinear",
    lon_dim: Optional[str] = None,
    lat_dim: Optional[str] = None,
) -> xarray.Dataset:
    """
    Convenience function to regrid a dataset.

    Args:
        dataset: Input xarray Dataset
        target_grid: Target grid for regridding
        method: Interpolation method ('nearest', 'bilinear', 'conservative')
        lon_dim: Name of longitude dimension (auto-detected if None)
        lat_dim: Name of latitude dimension (auto-detected if None)

    Returns:
        Regridded xarray Dataset
    """
    # Create source grid from dataset
    source_grid = Grid.from_dataset(dataset)

    # Select regridder based on method
    if method == "nearest":
        regridder = NearestRegridder(source_grid, target_grid)
    elif method == "bilinear":
        regridder = BilinearRegridder(source_grid, target_grid)
    elif method == "conservative":
        regridder = ConservativeRegridder(source_grid, target_grid)
    else:
        raise ValueError(
            f"Unknown method: {method}. Choose from 'nearest', 'bilinear', 'conservative'"
        )

    return regridder.regrid_dataset(dataset, lon_dim=lon_dim, lat_dim=lat_dim)



# skyborn/__init__.py
from .calculations import convert_longitude_range, linear_regression

from .gradients import (
    calculate_gradient,
    calculate_meridional_gradient,
    calculate_zonal_gradient,
    calculate_vertical_gradient,
)

from .causality import liang_causality, granger_causality

# Import conversion functions for easy access
from .conversion import (
    convert_grib_to_nc,
    convert_grib_to_nc_simple,
    batch_convert_grib_to_nc,
    grib2nc,
    grib_to_netcdf,
)

from . import plot
from . import interp
from . import ROF
from . import conversion

__version__ = "0.3.6"  # Updated to version 0.3.6



# import os
import numpy as np

# import pandas as pd
# from glob import glob
from scipy import stats
from typing import Dict, List, Tuple, Union, Any


def speco(C: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    This function computes eigenvalues and eigenvectors, in descending order
    :param C: numpy.ndarray
        A p x p symmetric real matrix
    :return:
    eigenvectors: numpy.ndarray
        The eigenvectors (eigenvectors[:, i] is the i-th eigenvector)
    eigenvalues_diag: numpy.ndarray
        The eigenvalues as a diagonal matrix
    """
    # Compute eigenvalues and eigenvectors (the eigenvectors are non unique so the values may change from one software
    # to another e.g. python, matlab, scilab)
    raw_eigenvals, raw_eigenvecs = np.linalg.eig(C)

    # Take real part (to avoid numeric noise, eg small complex numbers)
    if np.max(np.imag(raw_eigenvals)) / np.max(np.real(raw_eigenvals)) > 1e-12:
        raise ValueError("Matrix is not symmetric")

    # Check that C is symmetric (<=> real eigen-values/-vectors)
    eigenvecs_real = np.real(raw_eigenvecs)
    eigenvals_real = np.real(raw_eigenvals)

    # Sort eigenvalues in descending order and get their indices to order the eigenvectors
    eigenvals_sorted = np.sort(eigenvals_real)[::-1]
    sort_indices = np.argsort(eigenvals_real)[::-1]

    eigenvectors = eigenvecs_real[:, sort_indices]
    eigenvalues_diag = np.diag(eigenvals_sorted)

    return eigenvectors, eigenvalues_diag


def chi2_test(d_cons: float, df: int) -> float:
    """
    Check whether it is from a chi-squared distribution or not
    :param d_cons: float
        -2 log-likelihood
    :param df: int
        Degrees of freedom
    :return:
    pv_cons: float
        p-value for the test
    """
    rien = stats.chi2.cdf(d_cons, df=df)
    pv_cons = 1.0 - rien

    return pv_cons


def project_vectors(nt: int, X: np.ndarray) -> np.ndarray:
    """
    This function provides a projection matrix U that can be applied to X to ensure its covariance matrix to be
    full-ranked. Projects to a nt-1 subspace (ref: Ribes et al., 2013).
    :param nt: int
        number of time steps
    :param X: numpy.ndarray
        nt x nf array to be projected
    :return:
    projected_X: numpy.ndarray
        nt - 1 x nf array of projected timeseries
    """
    # Create centering matrix M = I - (1/nt) * ones_matrix
    # This removes the mean from each time series
    centering_matrix = np.eye(nt, nt) - np.ones((nt, nt)) / nt

    # Compute eigen-decomposition of centering matrix
    # Note: rank(M) = nt-1, so M has one eigenvalue equal to 0
    eigenvectors, _ = speco(centering_matrix)

    # Select first (nt-1) eigenvectors corresponding to non-zero eigenvalues
    # These form the projection matrix U
    projection_matrix = eigenvectors[:, : nt - 1].T

    # Apply projection to input data
    projected_X = np.dot(projection_matrix, X)

    return projected_X


def unproject_vectors(nt: int, Xc: np.ndarray) -> np.ndarray:
    """
    Unproject data from (nt-1) subspace back to nt subspace for trend computation.

    :param nt: int
        Number of time steps
    :param Xc: numpy.ndarray
        Projected data matrix (nt-1 x nf) to be unprojected
    :return:
    numpy.ndarray
        Unprojected data matrix (nt x nf)
    """
    # Create centering matrix: I - (1/nt)*ones
    centering_matrix = np.eye(nt, nt) - np.ones((nt, nt)) / nt

    # Get eigenvectors
    eigenvectors, _ = speco(centering_matrix)

    # Compute inverse projection matrix
    # Note: This is the pseudo-inverse for the reduced subspace
    inverse_projection = np.linalg.inv(eigenvectors.T)[:, : nt - 1]

    # Apply inverse projection
    unprojected_X = np.dot(inverse_projection, Xc)

    return unprojected_X


def SSM(X_dict: Dict[str, np.ndarray], X_mm: np.ndarray) -> np.ndarray:
    """
    Calculates the squared difference between each models ensemble mean and the multi-model mean. Based on
    (Ribes et al., 2017)
    :param X_dict: dict
        Dictionary where keys are experiment names and values are arrays (n_members, n_time)
    :param X_mm: numpy.ndarray
        Multi-model ensemble mean, shape (n_time,)
    :return:
    np.diag(((Xc - Xc_mm) ** 2.).sum(axis=1)): numpy.ndarray
        nt -1 x nt - 1 array of the difference between each model ensemble mean the multi-model mean
    """
    # Make sure X_mm has right shape
    if X_mm.ndim == 1:
        X_mm = X_mm.reshape((len(X_mm), 1))

    # Calculate ensemble mean for each experiment
    exp_means = []
    for exp_name, exp_data in X_dict.items():
        # Get ensemble mean for this experiment
        ensemble_mean = np.mean(exp_data, axis=0)  # shape: (n_time,)
        exp_means.append(ensemble_mean)

    # Stack all experiment means: (n_time, n_experiments)
    X = np.column_stack(exp_means)

    # Apply projection (this is the default behavior in original SSM)
    Xc = project_vectors(X.shape[0], X)
    Xc_mm = project_vectors(X.shape[0], X_mm)

    return np.diag(((Xc - Xc_mm) ** 2.0).sum(axis=1))


def get_nruns(X_dict: Dict[str, np.ndarray]) -> np.ndarray:
    """
    Gets the number of runs for each CESM2 experiment

    :param X_dict: dict
        Dictionary where keys are experiment names (e.g., 'CESM2-GHG', 'CESM2-EE')
        and values are arrays with shape (n_members, n_time)
    :return:
    nruns: numpy.ndarray
        Array with number of runs for each experiment
    """
    nruns = []
    for exp_name, exp_data in X_dict.items():
        nruns.append(exp_data.shape[0])  # number of members
    return np.array(nruns)


def Cm_estimate(
    X_dict: Dict[str, np.ndarray], Cv: np.ndarray, X_mm: np.ndarray
) -> np.ndarray:
    """
    Estimated covariance matrix for model error (Ribes et al., 2017)
    Modified for CESM2 experiments with ensemble member arrays

    :param X_dict: dict
        Dictionary where keys are experiment names (e.g., 'CESM2-GHG', 'CESM2-EE')
        and values are arrays with shape (n_members, n_time)
    :param Cv: numpy.ndarray
        Array with internal variability covariance matrix
    :param X_mm: numpy.ndarray
        Array with multi-model ensemble mean
    :return:
    Cm_pos_hat: numpy.ndarray
        Estimated covariance matrix for model error
    """

    # Calculate model differences using our modified SSM function
    _SSM = SSM(X_dict, X_mm)

    # Get number of runs and number of models using our modified function
    nruns = get_nruns(X_dict)
    nm = len(nruns)  # number of experiments

    # Calculate Cv_all based on number of runs for each experiment
    Cv_all = np.zeros(Cv.shape)
    for nr in nruns:
        Cv_all += Cv / nr

    # First estimation of Cm
    Cm_hat = (1.0 / (nm - 1.0)) * (_SSM - ((nm - 1.0) / nm) * Cv_all)

    # Set negative eigenvalues to zero and recompose the signal
    S, X = np.linalg.eig(Cm_hat)
    S[S < 0] = 0
    Cm_pos_hat = np.linalg.multi_dot(
        [X, np.diag(S), np.linalg.inv(X)]
    )  # spectral decomposition

    Cm_pos_hat = (1.0 + (1.0 / nm)) * Cm_pos_hat

    return Cm_pos_hat


def Cv_estimate(X_dict: Dict[str, np.ndarray], Cv: np.ndarray) -> np.ndarray:
    """
    Estimated covariance matrix for internal variability considering multiple models (Ribes et al., 2017)
    Modified for CESM2 experiments with ensemble member arrays

    :param X_dict: dict
        Dictionary where keys are experiment names and values are arrays (n_members, n_time)
    :param Cv: numpy.ndarray
        Array with internal variability covariance matrix
    :return:
    Cv_estimate: numpy.ndarray
        Estimated covariance matrix for internal variability considering multiple models
    """
    # Get number of runs and number of models
    nruns = get_nruns(X_dict)
    nm = len(nruns)

    Cv_all = np.zeros(Cv.shape)
    for nr in nruns:
        Cv_all += Cv / nr

    Cv_estimate = (1.0 / (nm**2.0)) * Cv_all

    return Cv_estimate


if __name__ == "__main__":
    T = 11
    M = np.eye(T, T) - np.ones((T, T)) / T
    speco(M)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Utilities for Liang and Granger causality analysis
https://github.com/LinkedEarth/Pyleoclim_util
"""

__all__ = ["liang_causality", "granger_causality"]
from statsmodels.tsa.arima_process import arma_generate_sample
from statsmodels.tsa.arima.model import ARIMA
import numpy as np
from statsmodels.tsa.stattools import grangercausalitytests
from tqdm import tqdm
from scipy.stats.mstats import mquantiles


def ar1_fit_evenly(y):
    """Returns the lag-1 autocorrelation from AR(1) fit.

    Uses `statsmodels.tsa.arima.model.ARIMA <https://www.statsmodels.org/devel/generated/statsmodels.tsa.arima.model.ARIMA.html>`_. to
    calculate lag-1 autocorrelation

    MARK FOR DEPRECATION once uar1_fit is adopted

    Parameters
    ----------
    y : array
        Vector of (float) numbers as a time series

    Returns
    -------
    g : float
        Lag-1 autocorrelation coefficient

    """
    # syntax compatible with statsmodels v0.11.1
    # ar1_mod = sm.tsa.ARMA(y, (1, 0), missing='drop').fit(trend='nc', disp=0)
    # syntax compatible with statsmodels v0.12
    ar1_mod = ARIMA(y, order=(1, 0, 0), missing="drop", trend="ct").fit()
    g = ar1_mod.params[2]

    if g > 1:
        print(
            "Warning: AR(1) fitted autocorrelation greater than 1; setting to 1-eps^{1/4}"
        )
        eps = np.spacing(1.0)
        g = 1.0 - eps ** (1 / 4)

    return g


def sm_ar1_sim(n, p, g, sig):
    """Produce p realizations of an AR1 process of length n with lag-1 autocorrelation g using statsmodels

    Parameters
    ----------
    n : int
        row dimensions
    p : int
        column dimensions

    g : float
        lag-1 autocorrelation coefficient

    sig : float
        the standard deviation of the original time series

    Returns
    -------
    red : numpy matrix
        n rows by p columns matrix of an AR1 process

    See also
    --------

    pyleoclim.utils.correlation.corr_sig : Estimates the Pearson's correlation and associated significance between two non IID time series
    pyleoclim.utils.correlation.fdr : Determine significance based on the false discovery rate

    """
    # specify model parameters (statsmodel wants lag0 coefficents as unity)
    ar = np.r_[1, -g]  # AR model parameter
    ma = np.r_[1, 0.0]  # MA model parameters
    # theoretical noise variance for red to achieve the same variance as X
    sig_n = sig * np.sqrt(1 - g**2)

    red = np.empty(shape=(n, p))  # declare array

    # simulate AR(1) model for each column
    for i in np.arange(p):
        red[:, i] = arma_generate_sample(
            ar=ar, ma=ma, nsample=n, burnin=50, scale=sig_n
        )

    return red


# -------
# Main functions
# --------


def granger_causality(y1, y2, maxlag=1, addconst=True, verbose=True):
    """Granger causality tests

    Four tests for the Granger non-causality of 2 time series.

    All four tests give similar results. params_ftest and ssr_ftest are equivalent based on F test which is identical to lmtest:grangertest in R.

    Wrapper for the functions described in statsmodels (https://www.statsmodels.org/stable/generated/statsmodels.tsa.stattools.grangercausalitytests.html)

    Parameters
    ----------
    y1, y2: array
        vectors of (real) numbers with identical length, no NaNs allowed
    maxlag : int or int iterable, optional
        If an integer, computes the test for all lags up to maxlag. If an iterable, computes the tests only for the lags in maxlag.
    addconst : bool, optional
        Include a constant in the model.
    verbose : bool, optional
        Print results

    Returns
    -------
    dict
        All test results, dictionary keys are the number of lags. For each lag the values are a tuple, with the first element a dictionary with test statistic,
        pvalues, degrees of freedom, the second element are the OLS estimation results for the restricted model, the unrestricted model and the restriction (contrast)
        matrix for the parameter f_test.

    Notes
    -----

    The null hypothesis for Granger causality tests is that y2, does NOT Granger cause y1. Granger causality means that past values of y2 have a statistically significant effect on the current value of y1, taking past values of y1 into account as regressors. We reject the null hypothesis that y2 does not Granger cause y1 if the p-values are below a desired threshold (e.g. 0.05).

    The null hypothesis for all four test is that the coefficients corresponding to past values of the second time series are zero.

    ‘params_ftest’, ‘ssr_ftest’ are based on the F distribution

    ‘ssr_chi2test’, ‘lrtest’ are based on the chi-square distribution

    See also
    --------

    pyleoclim.utils.causality.liang_causality : information flow estimated using the Liang algorithm

    pyleoclim.utils.causality.signif_isopersist : significance test with AR(1) with same persistence

    pyleoclim.utils.causality.signif_isospec : significance test with surrogates with randomized phases

    References
    ----------

    Granger, C. W. J. (1969). Investigating causal relations by econometric models and cross-spectral methods. Econometrica, 37(3), 424-438.

    Granger, C. W. J. (1980). Testing for causality: A personal viewpoont. Journal of Economic Dynamics and Control, 2, 329-352.

    Granger, C. W. J. (1988). Some recent development in a concept of causality. Journal of Econometrics, 39(1-2), 199-211.

    """

    if len(y1) != len(y2):
        raise ValueError("Timeseries must be of same length")

    x = np.array([y1, y2]).T
    res = grangercausalitytests(x, maxlag=maxlag, addconst=addconst, verbose=verbose)
    return res


def phaseran(recblk, nsurr):
    """Simultaneous phase randomization of a set of time series

    It creates blocks of surrogate data with the same second order properties as the original
    time series dataset by transforming the original data into the frequency domain, randomizing the
    phases simultaneoulsy across the time series and converting the data back into the time domain.

    Written by Carlos Gias for MATLAB

    http://www.mathworks.nl/matlabcentral/fileexchange/32621-phase-randomization/content/phaseran.m

    Parameters
    ----------
    recblk : numpy array
        2D array , Row: time sample. Column: recording.
        An odd number of time samples (height) is expected.
        If that is not the case, recblock is reduced by 1 sample before the surrogate data is created.
        The class must be double and it must be nonsparse.

    nsurr : int
        is the number of image block surrogates that you want to generate.

    Returns
    -------
    surrblk : numpy array
        3D multidimensional array image block with the surrogate datasey along the third dimension

    See also
    --------

    pyleoclim.utils.correlation.corr_sig : Estimates the Pearson's correlation and associated significance between two non IID time series
    pyleoclim.utils.correlation.fdf : Determine significance based on the false discovery rate

    References
    ----------

    - Prichard, D., Theiler, J. Generating Surrogate Data for Time Series with Several Simultaneously Measured Variables (1994) Physical Review Letters, Vol 73, Number 7

    - Carlos Gias (2020). Phase randomization, MATLAB Central File Exchange
    """
    # Get parameters
    nfrms = recblk.shape[0]

    if nfrms % 2 == 0:
        nfrms = nfrms - 1
        recblk = recblk[0:nfrms]

    len_ser = int((nfrms - 1) / 2)
    interv1 = np.arange(1, len_ser + 1)
    interv2 = np.arange(len_ser + 1, nfrms)

    # Fourier transform of the original dataset
    fft_recblk = np.fft.fft(recblk)

    surrblk = np.zeros((nfrms, nsurr))

    #  for k in tqdm(np.arange(nsurr)):
    for k in np.arange(nsurr):
        ph_rnd = np.random.rand(len_ser)

        # Create the random phases for all the time series
        ph_interv1 = np.exp(2 * np.pi * 1j * ph_rnd)
        ph_interv2 = np.conj(np.flipud(ph_interv1))

        # Randomize all the time series simultaneously
        fft_recblk_surr = np.copy(fft_recblk)
        fft_recblk_surr[interv1] = fft_recblk[interv1] * ph_interv1
        fft_recblk_surr[interv2] = fft_recblk[interv2] * ph_interv2

        # Inverse transform
        surrblk[:, k] = np.real(np.fft.ifft(fft_recblk_surr))

    return surrblk


def liang_causality(
    y1,
    y2,
    npt=1,
    signif_test="isospec",
    nsim=1000,
    qs=[0.005, 0.025, 0.05, 0.95, 0.975, 0.995],
):
    """Liang-Kleeman information flow

    Estimate the Liang information transfer from series y2 to series y1 with
    significance estimates using either an AR(1) tests with series with the same
    persistence or surrogates with randomized phases.

    Parameters
    ----------
    y1, y2 : array
        vectors of (real) numbers with identical length, no NaNs allowed
    npt : int >=1
        time advance in performing Euler forward differencing,
        e.g., 1, 2. Unless the series are generated with a highly chaotic deterministic system,
        npt=1 should be used
    signif_test : str; {'isopersist', 'isospec'}
        the method for significance test
        see signif_isospec and signif_isopersist for details.
    nsim : int
        the number of AR(1) surrogates for significance test
    qs : list
        the quantiles for significance test

    Returns
    -------
    res : dict
        A dictionary of results including:

        - T21 : float - information flow from y2 to y1 (Note: not y1 -> y2!)
        - tau21 : float - the standardized information flow from y2 to y1
        - Z : float - the total information flow from y2 to y1
        - dH1_star : float - dH*/dt (Liang, 2016)
        - dH1_noise : float
        - signif_qs : the quantiles for significance test
        - T21_noise : list - the quantiles of the information flow from noise2 to noise1 for significance testing
        - tau21_noise : list - the quantiles of the standardized information flow from noise2 to noise1 for significance testing

    See also
    --------
    pyleoclim.utils.causality.liang : information flow estimated using the Liang algorithm
    pyleoclim.utils.causality.granger_causality : information flow estimated using the Granger algorithm
    pyleoclim.utils.causality.signif_isopersist : significance test with AR(1) with same persistence
    pyleoclim.utils.causality.causality.signif_isospec : significance test with surrogates with randomized phases

    References
    ----------
    Liang, X.S. (2013) The Liang-Kleeman Information Flow: Theory and Applications. Entropy, 15, 327-360, doi:10.3390/e15010327

    Liang, X.S. (2014) Unraveling the cause-effect relation between timeseries. Physical review, E 90, 052150

    Liang, X.S. (2015) Normalizing the causality between time series. Physical review, E 92, 022126

    Liang, X.S. (2016) Information flow and causality as rigorous notions ab initio. Physical review, E 94, 052201

    """

    dt = 1
    nm = np.size(y1)

    grad1 = (y1[0 + npt :] - y1[0:-npt]) / (npt)
    grad2 = (y2[0 + npt :] - y2[0:-npt]) / (npt)

    y1 = y1[:-npt]
    y2 = y2[:-npt]

    N = nm - npt
    C = np.cov(y1, y2)
    detC = np.linalg.det(C)

    dC = np.ndarray((2, 2))
    dC[0, 0] = np.sum((y1 - np.mean(y1)) * (grad1 - np.mean(grad1)))
    dC[0, 1] = np.sum((y1 - np.mean(y1)) * (grad2 - np.mean(grad2)))
    dC[1, 0] = np.sum((y2 - np.mean(y2)) * (grad1 - np.mean(grad1)))
    dC[1, 1] = np.sum((y2 - np.mean(y2)) * (grad2 - np.mean(grad2)))

    dC /= N - 1

    a11 = C[1, 1] * dC[0, 0] - C[0, 1] * dC[1, 0]
    a12 = -C[0, 1] * dC[0, 0] + C[0, 0] * dC[1, 0]

    a11 /= detC
    a12 /= detC

    f1 = np.mean(grad1) - a11 * np.mean(y1) - a12 * np.mean(y2)
    R1 = grad1 - (f1 + a11 * y1 + a12 * y2)
    Q1 = np.sum(R1 * R1)
    b1 = np.sqrt(Q1 * dt / N)

    NI = np.ndarray((4, 4))
    NI[0, 0] = N * dt / b1**2
    NI[1, 1] = dt / b1**2 * np.sum(y1 * y1)
    NI[2, 2] = dt / b1**2 * np.sum(y2 * y2)
    NI[3, 3] = 3 * dt / b1**4 * np.sum(R1 * R1) - N / b1**2
    NI[0, 1] = dt / b1**2 * np.sum(y1)
    NI[0, 2] = dt / b1**2 * np.sum(y2)
    NI[0, 3] = 2 * dt / b1**3 * np.sum(R1)
    NI[1, 2] = dt / b1**2 * np.sum(y1 * y2)
    NI[1, 3] = 2 * dt / b1**3 * np.sum(R1 * y1)
    NI[2, 3] = 2 * dt / b1**3 * np.sum(R1 * y2)

    NI[1, 0] = NI[0, 1]
    NI[2, 0] = NI[0, 2]
    NI[2, 1] = NI[1, 2]
    NI[3, 0] = NI[0, 3]
    NI[3, 1] = NI[1, 3]
    NI[3, 2] = NI[2, 3]

    invNI = np.linalg.pinv(NI)
    var_a12 = invNI[2, 2]
    T21 = C[0, 1] / C[0, 0] * (-C[1, 0] * dC[0, 0] + C[0, 0] * dC[1, 0]) / detC
    var_T21 = (C[0, 1] / C[0, 0]) ** 2 * var_a12

    dH1_star = a11
    dH1_noise = b1**2 / (2 * C[0, 0])

    Z = np.abs(T21) + np.abs(dH1_star) + np.abs(dH1_noise)

    tau21 = T21 / Z
    dH1_star = dH1_star / Z
    dH1_noise = dH1_noise / Z

    signif_test_func = {
        "isopersist": signif_isopersist,
        "isospec": signif_isospec,
    }

    signif_dict = signif_test_func[signif_test](
        y1, y2, method="liang", nsim=nsim, qs=qs, npt=npt
    )
    T21_noise_qs = signif_dict["T21_noise_qs"]
    tau21_noise_qs = signif_dict["tau21_noise_qs"]

    res = {
        "T21": T21,
        "tau21": tau21,
        "Z": Z,
        "dH1_star": dH1_star,
        "dH1_noise": dH1_noise,
        "signif_qs": qs,
        "T21_noise": T21_noise_qs,
        "tau21_noise": tau21_noise_qs,
    }

    return res


def liang(y1, y2, npt=1):
    """
    Estimate the Liang information transfer from series y2 to series y1

    Parameters
    ----------
    y1, y2 : array
        Vectors of (real) numbers with identical length, no NaNs allowed

    npt : int  >=1
        Time advance in performing Euler forward differencing,
        e.g., 1, 2. Unless the series are generated with a highly chaotic deterministic system,
        npt=1 should be used

    Returns
    -------
    res : dict
        A dictionary of results including:

            - T21 (float): information flow from y2 to y1 (Note: not y1 -> y2!)
            - tau21 (float): the standardized information flow from y2 to y1
            - Z (float): the total information flow from y2 to y1
            - dH1_star (float): dH*/dt (Liang, 2016)
            - dH1_noise (float)

    See also
    --------

    pyleoclim.utils.causality.liang_causality : information flow estimated using the Liang algorithm
    pyleoclim.utils.causality.granger_causality : information flow estimated using the Granger algorithm
    pyleoclim.utils.causality.signif_isopersist : significance test with AR(1) with same persistence
    pyleoclim.utils.causality.signif_isospec : significance test with surrogates with randomized phases

    References
    ----------

    Liang, X.S. (2013) The Liang-Kleeman Information Flow: Theory and
            Applications. Entropy, 15, 327-360, doi:10.3390/e15010327

    Liang, X.S. (2014) Unraveling the cause-effect relation between timeseries.
        Physical review, E 90, 052150

    Liang, X.S. (2015) Normalizing the causality between time series.
        Physical review, E 92, 022126

    Liang, X.S. (2016) Information flow and causality as rigorous notions ab initio.
        Physical review, E 94, 052201

    """
    dt = 1
    nm = np.size(y1)

    grad1 = (y1[0 + npt :] - y1[0:-npt]) / (npt)
    grad2 = (y2[0 + npt :] - y2[0:-npt]) / (npt)

    y1 = y1[:-npt]
    y2 = y2[:-npt]

    N = nm - npt
    C = np.cov(y1, y2)
    detC = np.linalg.det(C)

    dC = np.ndarray((2, 2))
    dC[0, 0] = np.sum((y1 - np.mean(y1)) * (grad1 - np.mean(grad1)))
    dC[0, 1] = np.sum((y1 - np.mean(y1)) * (grad2 - np.mean(grad2)))
    dC[1, 0] = np.sum((y2 - np.mean(y2)) * (grad1 - np.mean(grad1)))
    dC[1, 1] = np.sum((y2 - np.mean(y2)) * (grad2 - np.mean(grad2)))

    dC /= N - 1

    a11 = C[1, 1] * dC[0, 0] - C[0, 1] * dC[1, 0]
    a12 = -C[0, 1] * dC[0, 0] + C[0, 0] * dC[1, 0]

    a11 /= detC
    a12 /= detC

    f1 = np.mean(grad1) - a11 * np.mean(y1) - a12 * np.mean(y2)
    R1 = grad1 - (f1 + a11 * y1 + a12 * y2)
    Q1 = np.sum(R1 * R1)
    b1 = np.sqrt(Q1 * dt / N)

    NI = np.ndarray((4, 4))
    NI[0, 0] = N * dt / b1**2
    NI[1, 1] = dt / b1**2 * np.sum(y1 * y1)
    NI[2, 2] = dt / b1**2 * np.sum(y2 * y2)
    NI[3, 3] = 3 * dt / b1**4 * np.sum(R1 * R1) - N / b1**2
    NI[0, 1] = dt / b1**2 * np.sum(y1)
    NI[0, 2] = dt / b1**2 * np.sum(y2)
    NI[0, 3] = 2 * dt / b1**3 * np.sum(R1)
    NI[1, 2] = dt / b1**2 * np.sum(y1 * y2)
    NI[1, 3] = 2 * dt / b1**3 * np.sum(R1 * y1)
    NI[2, 3] = 2 * dt / b1**3 * np.sum(R1 * y2)

    NI[1, 0] = NI[0, 1]
    NI[2, 0] = NI[0, 2]
    NI[2, 1] = NI[1, 2]
    NI[3, 0] = NI[0, 3]
    NI[3, 1] = NI[1, 3]
    NI[3, 2] = NI[2, 3]

    invNI = np.linalg.pinv(NI)
    var_a12 = invNI[2, 2]
    T21 = C[0, 1] / C[0, 0] * (-C[1, 0] * dC[0, 0] + C[0, 0] * dC[1, 0]) / detC
    var_T21 = (C[0, 1] / C[0, 0]) ** 2 * var_a12

    dH1_star = a11
    dH1_noise = b1**2 / (2 * C[0, 0])

    Z = np.abs(T21) + np.abs(dH1_star) + np.abs(dH1_noise)

    tau21 = T21 / Z
    dH1_star = dH1_star / Z
    dH1_noise = dH1_noise / Z

    res = {
        "T21": T21,
        "tau21": tau21,
        "Z": Z,
        "dH1_star": dH1_star,
        "dH1_noise": dH1_noise,
    }

    return res


def signif_isopersist(
    y1, y2, method, nsim=1000, qs=[0.005, 0.025, 0.05, 0.95, 0.975, 0.995], **kwargs
):
    """significance test with AR(1) with same persistence

    Parameters
    ----------
    y1, y2 : array
        vectors of (real) numbers with identical length, no NaNs allowed

    method : str; {'liang'}
        estimates for the Liang method

    nsim : int
        the number of AR(1) surrogates for significance test

    qs : list
        the quantiles for significance test

    Returns
    -------
    res_dict : dict

        A dictionary with the following information:

          - T21_noise_qs : list
            the quantiles of the information flow from noise2 to noise1 for significance testing
          - tau21_noise_qs : list
            the quantiles of the standardized information flow from noise2 to noise1 for significance testing

    See also
    --------

    pyleoclim.utils.causality.liang_causality : information flow estimated using the Liang algorithm
    pyleoclim.utils.causality.granger_causality : information flow estimated using the Granger algorithm
    pyleoclim.utils.causality.signif_isospec : significance test with surrogates with randomized phases

    """
    g1 = ar1_fit_evenly(y1)
    g2 = ar1_fit_evenly(y2)
    sig1 = np.std(y1)
    sig2 = np.std(y2)
    n = np.size(y1)
    noise1 = sm_ar1_sim(n, nsim, g1, sig1)
    noise2 = sm_ar1_sim(n, nsim, g2, sig2)

    if method == "liang":
        npt = kwargs["npt"] if "npt" in kwargs else 1
        T21_noise = []
        tau21_noise = []
        for i in tqdm(range(nsim), desc="Calculating causality between surrogates"):
            res_noise = liang(noise1[:, i], noise2[:, i], npt=npt)
            tau21_noise.append(res_noise["tau21"])
            T21_noise.append(res_noise["T21"])
        tau21_noise = np.array(tau21_noise)
        T21_noise = np.array(T21_noise)
        tau21_noise_qs = mquantiles(tau21_noise, qs)
        T21_noise_qs = mquantiles(T21_noise, qs)

        res_dict = {
            "tau21_noise_qs": tau21_noise_qs,
            "T21_noise_qs": T21_noise_qs,
        }
    # TODO add granger method
    else:
        raise KeyError(f"{method} is not a valid method")

    return res_dict


def signif_isospec(
    y1, y2, method, nsim=1000, qs=[0.005, 0.025, 0.05, 0.95, 0.975, 0.995], **kwargs
):
    """significance test with surrogates with randomized phases

    Parameters
    ----------
    y1, y2 : array
        vectors of (real) numbers with identical length, no NaNs allowed
    method : str; {'liang'}
        estimates for the Liang method
    nsim : int
        the number of surrogates for significance test
    qs : list
        the quantiles for significance test
    kwargs : dict
        keyword arguments for the causality method (e.g. npt for Liang-Kleeman)

    Returns
    -------
    res_dict : dict
        A dictionary with the following information:
          - T21_noise_qs : list
                        the quantiles of the information flow from noise2 to noise1 for significance testing
          - tau21_noise_qs : list
                          the quantiles of the standardized information flow from noise2 to noise1 for significance testing

    See also
    --------

    pyleoclim.utils.causality.liang_causality : information flow estimated using the Liang algorithm
    pyleoclim.utils.causality.granger_causality : information flow estimated using the Granger algorithm
    pyleoclim.utils.causality.signif_isopersist : significance test with AR(1) with same persistence

    """

    noise1 = phaseran(y1, nsim)
    noise2 = phaseran(y2, nsim)

    if method == "liang":
        npt = kwargs["npt"] if "npt" in kwargs else 1
        T21_noise = []
        tau21_noise = []
        for i in tqdm(range(nsim), desc="Calculating causality between surrogates"):
            res_noise = liang(noise1[:, i], noise2[:, i], npt=npt)
            tau21_noise.append(res_noise["tau21"])
            T21_noise.append(res_noise["T21"])
        tau21_noise = np.array(tau21_noise)
        T21_noise = np.array(T21_noise)
        tau21_noise_qs = mquantiles(tau21_noise, qs)
        T21_noise_qs = mquantiles(T21_noise, qs)

        res_dict = {
            "tau21_noise_qs": tau21_noise_qs,
            "T21_noise_qs": T21_noise_qs,
        }
    # TODO Recode with Surrogate class
    else:
        raise KeyError(f"{method} is not a valid method")

    return res_dict



import matplotlib.pyplot as plt
import matplotlib.transforms as mtransforms
import numpy as np


def add_equal_axes(ax, loc, pad, width):
    """
    Add a new Axes with equal height or width next to the original Axes and return that object.

    Parameters
    ----------
    ax : Axes or array_like of Axes
        The original Axes, or can be an array of Axes.

    loc : {'left', 'right', 'bottom', 'top'}
        Position of the new Axes relative to the old Axes.

    pad : float
        Spacing between the new Axes and the old Axes.

    width: float
        When loc='left' or 'right', width represents the width of the new Axes.
        When loc='bottom' or 'top', width represents the height of the new Axes.

    Returns
    -------
    ax_new : Axes
        New Axes object.
    """
    # Whether ax is a single Axes or a group of Axes, get the size and position of ax.
    axes = np.atleast_1d(ax).ravel()
    bbox = mtransforms.Bbox.union([ax.get_position() for ax in axes])

    # Determine the size and position of the new Axes.
    if loc == "left":
        x0_new = bbox.x0 - pad - width
        x1_new = x0_new + width
        y0_new = bbox.y0
        y1_new = bbox.y1
    elif loc == "right":
        x0_new = bbox.x1 + pad
        x1_new = x0_new + width
        y0_new = bbox.y0
        y1_new = bbox.y1
    elif loc == "bottom":
        x0_new = bbox.x0
        x1_new = bbox.x1
        y0_new = bbox.y0 - pad - width
        y1_new = y0_new + width
    elif loc == "top":
        x0_new = bbox.x0
        x1_new = bbox.x1
        y0_new = bbox.y1 + pad
        y1_new = y0_new + width

    # Create new Axes.
    fig = axes[0].get_figure()
    bbox_new = mtransforms.Bbox.from_extents(x0_new, y0_new, x1_new, y1_new)
    ax_new = fig.add_axes(bbox_new)

    return ax_new


def createFigure(figsize=(12, 8), dpi=300, subplotAdj=None, **kwargs):
    figsize = figsize
    figure = plt.figure(figsize=figsize, dpi=dpi, **kwargs)
    if subplotAdj is not None:
        plt.subplots_adjust(**subplotAdj)
    return figure



"""
Functions for curved quiver plots.
"""

from __future__ import annotations
from collections.abc import Hashable
from typing import TYPE_CHECKING

# import matplotlib.font_manager
from .modplot import CurvedQuiverplotSet
from typing import Literal
from matplotlib.patches import FancyArrowPatch

import xarray as xr
import matplotlib.pyplot as plt

# import numpy as np
# import matplotlib
from matplotlib.artist import Artist, allow_rasterization
from matplotlib.patches import Rectangle
from matplotlib.text import Text
from matplotlib.figure import Figure
from matplotlib.backend_bases import RendererBase

if TYPE_CHECKING:
    from matplotlib.axes import Axes

__all__ = ["curved_quiver", "add_curved_quiverkey"]


def curved_quiver(
    ds: xr.Dataset,
    x: Hashable,
    y: Hashable,
    u: Hashable,
    v: Hashable,
    ax: Axes | None = None,
    density=1,
    linewidth=None,
    color=None,
    cmap=None,
    norm=None,
    arrowsize=1,
    arrowstyle="-|>",
    transform=None,
    zorder=None,
    start_points=None,
    integration_direction="both",
    grains=15,
    broken_streamlines=True,
) -> CurvedQuiverplotSet:
    """
    Plot streamlines of a vector flow.

    .. warning::

        This function is experimental and the API is subject to change. Please use with caution.

    Parameters
    ----------
    ds : :py:class:`xarray.Dataset`.
        Wind dataset.
    x : Hashable or None, optional.
        Variable name for x-axis.
    y : Hashable or None, optional.
        Variable name for y-axis.
    u : Hashable or None, optional.
        Variable name for the u velocity (in `x` direction).
    v : Hashable or None, optional.
        Variable name for the v velocity (in `y` direction).
    ax : :py:class:`matplotlib.axes.Axes`, optional.
        Axes on which to plot. By default, use the current axes. Mutually exclusive with `size` and `figsize`.
    density : float or (float, float)
        Controls the closeness of streamlines. When ``density = 1``, the domain
        is divided into a 30x30 grid. *density* linearly scales this grid.
        Each cell in the grid can have, at most, one traversing streamline.
        For different densities in each direction, use a tuple
        (density_x, density_y).
    linewidth : float or 2D array
        The width of the streamlines. With a 2D array the line width can be
        varied across the grid. The array must have the same shape as *u*
        and *v*.
    color : color or 2D array
        The streamline color. If given an array, its values are converted to
        colors using *cmap* and *norm*.  The array must have the same shape
        as *u* and *v*.
    cmap, norm
        Data normalization and colormapping parameters for *color*; only used
        if *color* is an array of floats. See `~.Axes.imshow` for a detailed
        description.
    arrowsize : float
        Scaling factor for the arrow size.
    arrowstyle : str
        Arrow style specification.
        See `~matplotlib.patches.FancyArrowPatch`.
    start_points : (N, 2) array
        Coordinates of starting points for the streamlines in data coordinates
        (the same coordinates as the *x* and *y* arrays).
    zorder : float
        The zorder of the streamlines and arrows.
        Artists with lower zorder values are drawn first.
    integration_direction : {'forward', 'backward', 'both'}, default: 'both'
        Integrate the streamline in forward, backward or both directions.
    broken_streamlines : boolean, default: True
        If False, forces streamlines to continue until they
        leave the plot domain.  If True, they may be terminated if they
        come too close to another streamline.

    Returns
    -------
    CurvedQuiverplotSet
        Container object with attributes

        - ``lines``: `.LineCollection` of streamlines

        - ``arrows``: `.PatchCollection` containing `.FancyArrowPatch`
          objects representing the arrows half-way along streamlines.

            This container will probably change in the future to allow changes
            to the colormap, alpha, etc. for both lines and arrows, but these
            changes should be backward compatible.

    .. seealso::
        - https://github.com/matplotlib/matplotlib/issues/20038
        - https://github.com/kieranmrhunt/curved-quivers
        - https://github.com/Deltares/dfm_tools/issues/483
        - https://github.com/NCAR/geocat-viz/issues/4
        - https://docs.xarray.dev/en/stable/generated/xarray.Dataset.plot.streamplot.html#xarray.Dataset.plot.streamplot
    """
    from .modplot import velovect

    ds = ds.sortby(y)
    x = ds[x].data
    y = ds[y].data
    u = ds[u].data
    v = ds[v].data

    # https://scitools.org.uk/cartopy/docs/latest/gallery/miscellanea/logo.html#sphx-glr-gallery-miscellanea-logo-py
    if ax is None:
        ax = plt.gca()
    if type(transform).__name__ == "PlateCarree":
        transform = transform._as_mpl_transform(ax)

    # https://github.com/Deltares/dfm_tools/issues/294
    # https://github.com/Deltares/dfm_tools/blob/main/dfm_tools/modplot.py
    obj = velovect(
        ax,
        x,
        y,
        u,
        v,
        density=density,
        linewidth=linewidth,
        color=color,
        cmap=cmap,
        norm=norm,
        arrowsize=arrowsize,
        arrowstyle=arrowstyle,
        transform=transform,
        zorder=zorder,
        start_points=start_points,
        integration_direction=integration_direction,
        grains=grains,
        broken_streamlines=broken_streamlines,
    )
    return obj


class CurvedQuiverLegend(Artist):
    """Curved quiver legend with white background box, arrows scaled according to actual wind speed"""

    def __init__(
        self,
        ax,
        curved_quiver_set,  # Pass the original curved quiver object
        U: float,
        units: str = "m/s",
        width: float = 0.15,
        height: float = 0.08,
        loc: Literal[
            "lower left", "lower right", "upper left", "upper right"
        ] = "lower right",
        # Added labelpos parameter
        labelpos: Literal["N", "S", "E", "W"] = "E",
        max_arrow_length: float = 0.08,  # Maximum arrow length
        arrow_props: dict = None,
        patch_props: dict = None,
        text_props: dict = None,
        padding: float = 0.01,
        margin: float = 0.02,
        reference_speed: float = 2.0,
        # If None, automatically determined based on whether units is empty
        center_label: bool = None,
    ) -> None:
        """
        Initialize curved quiver legend with arrows proportional to wind speed

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes to add the legend to
        curved_quiver_set : CurvedQuiverplotSet
            Original curved quiver object, used to get scaling information
        U : float
            Wind speed value represented by the arrow
        units : str
            Unit string, if empty string(""), units won't be displayed and label will be auto-centered
        labelpos : {'N', 'S', 'E', 'W'}, default: 'E'
            Label position relative to arrow:
            'N' - Label above the arrow
            'S' - Label below the arrow
            'E' - Label to the right of the arrow
            'W' - Label to the left of the arrow
        center_label : bool or None
            Whether to center the label. If None, auto-center when units is empty
        max_arrow_length : float
            Maximum arrow length (relative to legend box)
        """
        super().__init__()
        self.reference_speed = reference_speed
        self.margin = margin
        self.ax = ax
        self.curved_quiver_set = curved_quiver_set
        self.U = U
        self.units = units
        self.labelpos = labelpos

        # Automatically determine whether to center label based on units
        self.show_units = units != ""
        if center_label is None:
            self.center_label = not self.show_units  # Auto-center if no units
        else:
            self.center_label = center_label

        self.width = width
        self.height = height
        self.loc = loc
        self.max_arrow_length = max_arrow_length
        self.padding = padding

        # Set default properties
        self.arrow_props = self._setup_arrow_props(arrow_props)
        self.patch_props = self._setup_patch_props(patch_props)
        self.text_props = self._setup_text_props(text_props)

        # Calculate actual arrow length (based on wind speed scale)
        self.arrow_length = self._calculate_arrow_length()

        # Create text content
        if self.center_label:
            self.text_content = f"{U}"
        else:
            self.text_content = f"{U}" if not self.show_units else f"{U} {units}"

        # Create temporary text object to measure size
        import matplotlib.pyplot as plt

        fig = plt.figure(figsize=(10, 10))  # Create temporary figure
        renderer = fig.canvas.get_renderer()

        temp_text = Text(0, 0, self.text_content, **self.text_props)
        temp_text.set_figure(fig)
        bbox = temp_text.get_window_extent(renderer)
        text_width = bbox.width / fig.dpi / 10  # 10 is the width of temporary figure
        text_height = bbox.height / fig.dpi / 10

        plt.close(fig)  # Close temporary figure

        # Adjust box size to ensure it can contain text and arrow
        # Adjust for different labelpos
        if labelpos in ["E", "W"]:  # Horizontal layout
            if self.center_label:
                # Center mode
                min_width = self.arrow_length + 3 * self.padding + text_width
                total_content_width = self.arrow_length + self.padding + text_width
            else:
                # Adjust based on labelpos
                if labelpos == "E":  # Text to the right of the arrow
                    min_width = (
                        2 * self.padding
                        + self.arrow_length
                        + self.padding
                        + text_width
                        + self.padding
                    )
                else:  # labelpos == 'W', text to the left of the arrow
                    min_width = (
                        2 * self.padding
                        + text_width
                        + self.padding
                        + self.arrow_length
                        + self.padding
                    )
                total_content_width = min_width - 2 * self.padding

            # Ensure box is wide enough
            if min_width > self.width:
                self.width = min_width

            # Adjust height to fit text and arrow
            min_height = max(text_height, 0.03) + 2 * self.padding

        else:  # Vertical layout (labelpos in ['N', 'S'])
            # In vertical direction, text and arrow are stacked
            min_height = (
                text_height + 2 * self.padding + 0.03
            )  # 0.03 is an estimate for arrow height

            # Ensure box is wide enough to accommodate the maximum width of arrow and text
            min_width = max(self.arrow_length, text_width) + 2 * self.padding

            if min_width > self.width:
                self.width = min_width

        # Ensure box height is sufficient
        if min_height > self.height:
            self.height = min_height

        # Calculate bottom-left coordinates of legend box based on position
        self._calculate_position()

        # Create background box
        self.patch = Rectangle(
            xy=(self.x, self.y),
            width=self.width,
            height=self.height,
            transform=ax.transAxes,
            **self.patch_props,
        )

        # Set arrow and text positions based on labelpos
        self._position_arrow_and_text(text_width, text_height)

        # Set z-order
        self.set_zorder(10)
        self.patch.set_zorder(9)
        self.arrow.set_zorder(11)
        self.text.set_zorder(11)

        # Add to axes
        ax.add_artist(self.patch)
        ax.add_artist(self.arrow)
        ax.add_artist(self.text)
        ax.add_artist(self)

    def _position_arrow_and_text(self, text_width, text_height):
        """Position arrow and text based on labelpos"""
        box_center_x = self.x + self.width / 2
        box_center_y = self.y + self.height / 2

        # Set default text alignment
        self.text_props.update(
            {"verticalalignment": "center", "horizontalalignment": "center"}
        )

        if self.center_label:
            # Center mode - entire combination (arrow+text) is centered
            if self.labelpos in ["E", "W"]:  # Horizontal layout
                # Calculate total content width
                total_content_width = self.arrow_length + self.padding + text_width

                # Group start position to center the entire element
                group_start_x = box_center_x - (total_content_width / 2)

                if self.labelpos == "E":  # Text to the right of arrow
                    # Set arrow position
                    arrow_start_x = group_start_x
                    arrow_end_x = arrow_start_x + self.arrow_length

                    # Create arrow
                    self.arrow = FancyArrowPatch(
                        (arrow_start_x, box_center_y),
                        (arrow_end_x, box_center_y),
                        transform=self.ax.transAxes,
                        **self.arrow_props,
                    )

                    # Set text position
                    text_x = arrow_end_x + self.padding
                    self.text_props["horizontalalignment"] = "left"

                else:  # self.labelpos == 'W', text to the left of arrow
                    # Place text first
                    text_x = group_start_x
                    self.text_props["horizontalalignment"] = "right"

                    # Set arrow position
                    arrow_start_x = text_x + text_width + self.padding
                    arrow_end_x = arrow_start_x + self.arrow_length

                    # Create arrow
                    self.arrow = FancyArrowPatch(
                        (arrow_start_x, box_center_y),
                        (arrow_end_x, box_center_y),
                        transform=self.ax.transAxes,
                        **self.arrow_props,
                    )

                # Create text
                self.text = Text(
                    text_x,
                    box_center_y,
                    self.text_content,
                    transform=self.ax.transAxes,
                    **self.text_props,
                )

            else:  # Vertical layout (self.labelpos in ['N', 'S'])
                if self.labelpos == "N":  # Text above the arrow
                    # Set arrow position (horizontally centered)
                    arrow_start_x = box_center_x - self.arrow_length / 2
                    arrow_end_x = box_center_x + self.arrow_length / 2

                    # Vertical position (lower half)
                    arrow_y = box_center_y - text_height / 2 - self.padding / 2

                    # Create arrow
                    self.arrow = FancyArrowPatch(
                        (arrow_start_x, arrow_y),
                        (arrow_end_x, arrow_y),
                        transform=self.ax.transAxes,
                        **self.arrow_props,
                    )

                    # Set text position (upper half)
                    text_y = box_center_y + self.padding / 2
                    self.text_props["verticalalignment"] = "bottom"

                else:  # self.labelpos == 'S', text below the arrow
                    # Set arrow position (horizontally centered)
                    arrow_start_x = box_center_x - self.arrow_length / 2
                    arrow_end_x = box_center_x + self.arrow_length / 2

                    # Vertical position (upper half)
                    arrow_y = box_center_y + text_height / 2 + self.padding / 2

                    # Create arrow
                    self.arrow = FancyArrowPatch(
                        (arrow_start_x, arrow_y),
                        (arrow_end_x, arrow_y),
                        transform=self.ax.transAxes,
                        **self.arrow_props,
                    )

                    # Set text position (lower half)
                    text_y = box_center_y - self.padding / 2
                    self.text_props["verticalalignment"] = "top"

                # Create text (horizontally centered)
                self.text = Text(
                    box_center_x,
                    text_y,
                    self.text_content,
                    transform=self.ax.transAxes,
                    **self.text_props,
                )

        else:  # Non-centered mode
            if self.labelpos in ["E", "W"]:  # Horizontal layout
                if self.labelpos == "E":  # Text to the right of arrow
                    # Layout starting from left
                    arrow_start_x = self.x + self.padding
                    arrow_end_x = arrow_start_x + self.arrow_length

                    # Create arrow
                    self.arrow = FancyArrowPatch(
                        (arrow_start_x, box_center_y),
                        (arrow_end_x, box_center_y),
                        transform=self.ax.transAxes,
                        **self.arrow_props,
                    )

                    # Create text label
                    text_x = arrow_end_x + self.padding
                    self.text_props["horizontalalignment"] = "left"

                else:  # self.labelpos == 'W', text to the left of arrow
                    # Layout starting from right
                    arrow_end_x = self.x + self.width - self.padding
                    arrow_start_x = arrow_end_x - self.arrow_length

                    # Create arrow
                    self.arrow = FancyArrowPatch(
                        (arrow_start_x, box_center_y),
                        (arrow_end_x, box_center_y),
                        transform=self.ax.transAxes,
                        **self.arrow_props,
                    )

                    # Create text label
                    text_x = arrow_start_x - self.padding
                    self.text_props["horizontalalignment"] = "right"

                self.text = Text(
                    text_x,
                    box_center_y,
                    self.text_content,
                    transform=self.ax.transAxes,
                    **self.text_props,
                )

            else:  # Vertical layout (self.labelpos in ['N', 'S'])
                # Horizontally center the arrow
                arrow_start_x = self.x + (self.width - self.arrow_length) / 2
                arrow_end_x = arrow_start_x + self.arrow_length

                if self.labelpos == "N":  # Text above the arrow
                    # Arrow at bottom
                    arrow_y = self.y + self.padding + 0.015  # Give the arrow some space

                    # Create arrow
                    self.arrow = FancyArrowPatch(
                        (arrow_start_x, arrow_y),
                        (arrow_end_x, arrow_y),
                        transform=self.ax.transAxes,
                        **self.arrow_props,
                    )

                    # Text at top
                    text_y = self.y + self.height - self.padding
                    self.text_props["verticalalignment"] = "top"

                else:  # self.labelpos == 'S', text below the arrow
                    # Arrow at top
                    arrow_y = self.y + self.height - self.padding - 0.015

                    # Create arrow
                    self.arrow = FancyArrowPatch(
                        (arrow_start_x, arrow_y),
                        (arrow_end_x, arrow_y),
                        transform=self.ax.transAxes,
                        **self.arrow_props,
                    )

                    # Text at bottom
                    text_y = self.y + self.padding
                    self.text_props["verticalalignment"] = "bottom"

                # Create text (horizontally centered)
                self.text = Text(
                    box_center_x,
                    text_y,
                    self.text_content,
                    transform=self.ax.transAxes,
                    **self.text_props,
                )

    def _calculate_arrow_length(self):
        """Calculate arrow length based on actual wind speed and original data"""
        try:
            # Use the scale information from the original curved_quiver
            if hasattr(self.curved_quiver_set, "resolution") and hasattr(
                self.curved_quiver_set, "magnitude"
            ):
                # resolution = self.curved_quiver_set.resolution
                # magnitude = self.curved_quiver_set.magnitude

                # Use reference speed to scale the arrow
                reference_speed = getattr(
                    self, "reference_speed", 2.0
                )  # Default reference speed is 2.0

                # Calculate the scale factor
                scale_factor = self.U / reference_speed

                # Ensure arrow length is within reasonable range
                arrow_length = min(
                    # Allow arrows up to 4x maximum
                    scale_factor * self.max_arrow_length,
                    self.max_arrow_length * 4,
                )
                arrow_length = max(
                    arrow_length, self.max_arrow_length * 0.2
                )  # Minimum length
                return arrow_length

            # Method 2: If scale information is unavailable, use simple linear scaling
            # Assume common wind speed range is 0-20 m/s
            scale_factor = min(self.U / 20.0, 1.0)
            arrow_length = max(
                scale_factor * self.max_arrow_length, self.max_arrow_length * 0.2
            )
            return arrow_length

        except Exception as e:
            print(f"Warning: Could not calculate proportional arrow length: {e}")
            # Fall back to fixed length
            return self.max_arrow_length * 0.6

    def _calculate_position(self):
        """Calculate legend box position based on loc parameter"""
        margin = getattr(self, "margin", 0.02)  # Use class attribute or default

        if self.loc == "lower left":
            self.x = margin
            self.y = margin
        elif self.loc == "lower right":
            self.x = 1 - self.width - margin
            self.y = margin
        elif self.loc == "upper left":
            self.x = margin
            self.y = 1 - self.height - margin
        elif self.loc == "upper right":
            self.x = 1 - self.width - margin
            self.y = 1 - self.height - margin
        else:
            raise ValueError(
                f"loc must be one of ['lower left', 'lower right', 'upper left', 'upper right'], got {self.loc}"
            )

    def _setup_arrow_props(self, arrow_props):
        """Set default arrow properties"""
        defaults = {
            "arrowstyle": "->",
            "mutation_scale": 20,
            "linewidth": 1.5,
            "color": "black",
        }
        if arrow_props:
            defaults.update(arrow_props)
        return defaults

    def _setup_patch_props(self, patch_props):
        """Set default background box properties"""
        defaults = {
            "linewidth": 1,
            "edgecolor": "black",
            "facecolor": "white",
            "alpha": 0.9,
        }
        if patch_props:
            defaults.update(patch_props)
        return defaults

    def _setup_text_props(self, text_props):
        """Set default text properties"""
        defaults = {
            "fontsize": 10,
            "verticalalignment": "center",
            "horizontalalignment": "left",
            "color": "black",
        }
        if text_props:
            defaults.update(text_props)
        return defaults

    def set_figure(self, fig: Figure) -> None:
        """Set figure object"""
        super().set_figure(fig)
        self.patch.set_figure(fig)
        self.arrow.set_figure(fig)
        self.text.set_figure(fig)

    @allow_rasterization
    def draw(self, renderer: RendererBase) -> None:
        """Draw the legend"""
        if self.get_visible():
            # Ensure all components are drawn
            self.patch.draw(renderer)
            self.arrow.draw(renderer)
            self.text.draw(renderer)
            self.stale = False


def add_curved_quiverkey(
    ax,
    curved_quiver_set,  # Pass the curved_quiver return object
    U: float = 2.0,
    units: str = "m/s",
    loc: str = "lower right",
    labelpos: str = "E",  # label position
    **kwargs,
) -> CurvedQuiverLegend:
    """
    Convenience function: Add proportionally scaled curved quiver legend to axes

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes object
    curved_quiver_set : CurvedQuiverplotSet
        Object returned by curved_quiver function
    U : float
        Wind speed value represented by the arrow
    units : str
        Unit
    loc : str
        Position
    labelpos : {'N', 'S', 'E', 'W'}, default: 'E'
        Label position relative to arrow:
        'N' - Label above the arrow
        'S' - Label below the arrow
        'E' - Label to the right of the arrow
        'W' - Label to the left of the arrow
    **kwargs
        Other parameters passed to CurvedQuiverLegend

    Returns
    -------
    CurvedQuiverLegend
        Legend object
    """
    return CurvedQuiverLegend(
        ax, curved_quiver_set, U, units=units, loc=loc, labelpos=labelpos, **kwargs
    )



import numpy as np
import xarray as xr
import metpy.calc as mpcalc
from metpy.units import units
from typing import Tuple, Union
from sklearn.feature_selection import f_regression


def linear_regression(
    data: Union[np.ndarray, xr.DataArray], predictor: Union[np.ndarray, xr.DataArray]
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Perform linear regression between a 3D data array and a predictor sequence.
    Handles both numpy arrays and xarray DataArrays.

    Parameters
    ----------
    data : np.ndarray or xr.DataArray
        A 3D array of shape (n_samples, dim1, dim2) containing dependent variables.
    predictor : np.ndarray or xr.DataArray
        A 1D array of shape (n_samples,) containing the independent variable.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        A tuple containing:
        - regression_coefficients: The slope of the regression line with shape (dim1, dim2)
        - p_values: The p-values of the regression with shape (dim1, dim2)

    Raises
    ------
    ValueError
        If the number of samples in data doesn't match the length of the predictor.
    """
    # Extract numpy arrays regardless of input type
    data = getattr(data, "values", data)
    predictor = getattr(predictor, "values", predictor)

    if len(data) != len(predictor):
        raise ValueError(
            f"Number of samples in data ({data.shape[0]}) must match "
            f"length of predictor ({len(predictor)})"
        )

    # Create design matrix with predictor and intercept
    design_matrix = np.column_stack((predictor, np.ones(predictor.shape[0])))

    # Get original dimensions and reshape for regression
    n_samples, dim1, dim2 = data.shape
    data_flat = data.reshape((n_samples, dim1 * dim2))

    # Perform linear regression
    regression_coef = np.linalg.lstsq(design_matrix, data_flat, rcond=None)[0][0]

    # Reshape results back to original dimensions
    regression_coef = regression_coef.reshape((dim1, dim2))

    # Calculate p-values
    p_values = f_regression(data_flat, predictor)[1].reshape(dim1, dim2)

    return regression_coef, p_values


def convert_longitude_range(
    data: Union[xr.DataArray, xr.Dataset], lon: str = "lon", center_on_180: bool = True
) -> Union[xr.DataArray, xr.Dataset]:
    """
    Wrap longitude coordinates of DataArray or Dataset to either -180..179 or 0..359.

    Parameters
    ----------
    data : xr.DataArray or xr.Dataset
        An xarray DataArray or Dataset object containing longitude coordinates.
    lon : str, optional
        The name of the longitude coordinate, default is 'lon'.
    center_on_180 : bool, optional
        If True, wrap longitude from 0..359 to -180..179;
        If False, wrap longitude from -180..179 to 0..359.

    Returns
    -------
    xr.DataArray or xr.Dataset
        The DataArray or Dataset with wrapped longitude coordinates.
    """
    # Wrap -180..179 to 0..359
    if center_on_180:
        data = data.assign_coords(**{lon: (lambda x: (x[lon] % 360))})
    # Wrap 0..359 to -180..179
    else:
        data = data.assign_coords(**{lon: (lambda x: ((x[lon] + 180) % 360) - 180)})
    return data.sortby(lon, ascending=True)



import numpy as np
from typing import Tuple, Literal
from .utils import speco


class PreProcess:
    """
    A class to preprocess model data to be used in attribution methods like OLS, TLS and others. The script is
    heavily based on Aurelien Ribes (CNRM-GAME) scilab code, so for more information the user should consult the
    following reference:
        Ribes A., S. Planton, L. Terray (2012) Regularised optimal fingerprint for attribution.
        Part I: method, properties and idealised analysis. Climate Dynamics.

    :attribute y: numpy.ndarray
        Array of size nt with observations as a timeseries
    :attribute X: numpy.ndarray
        Array of size nt x nf, where nf is the number of forcings, with the model output as a timeseries
    :attribute Z: numpy.ndarray
        Array of ensembles with internal variability used to compute covariance matrix
    :attribute nt: int
        Number of timesteps for the timeseries

    :method extract_Z2(self, method='regular', frac=0.5):
        Split big sample Z in Z1 and Z2
    :method proj_fullrank(self, Z1, Z2):
        Provides a projection matrix to ensure its covariance matrix to be full-ranked
    :method creg(self, X, method='ledoit', alpha1=1e-10, alpha2=1):
        Computes regularized covariance matrix
    """

    def __init__(self, y: np.ndarray, X: np.ndarray, Z: np.ndarray):
        """
        :param y: numpy.ndarray
            Array of size nt with observations as a timeseries
        :param X: numpy.ndarray
            Array of size nt x nf, where nf is the number of forcings, with the model output as a timeseries
        :param Z: numpy.ndarray
            Array of ensembles with internal variability used to compute covariance matrix
        """
        self.y = y
        self.X = X
        self.Z = Z

        self.nt = y.shape[0]

    def extract_Z2(
        self, method: Literal["regular"] = "regular", frac: float = 0.5
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        This function is used to split a big sample Z (dimension: nz x p, containing nz iid realisation of a random
        vector of size p) into two samples Z1 and Z2 (respectively of dimension nz1 x p and nz2 x p, with
        nz = nz1 + nz2). Further explanations in Ribes et al. (2012).
        :param method: str
            type of sampling used, for now may be only 'regular'
        :param frac: float
            fraction of realizations to put in Z2, the remaining is used in Z1
        :return:
        Z1: numpy.ndarray
            Array of size (nz1 x p)
        Z2: numpy.ndarray
            Array of size (nz2 x p)
        """
        nz = self.Z.shape[1]
        ind_z2 = np.zeros(nz)

        if method == "regular":
            # if frac = 0.5 ix would be [1, 3, 5, ..., ], so gets the index
            # for every two points
            # -1 is because python starts index at 0
            ix = np.arange(1 / frac - 1, nz, 1 / frac).astype(int)
            ind_z2[ix] = 1
            Z2 = self.Z[:, ind_z2 == 1]
            Z1 = self.Z[:, ind_z2 == 0]
        else:
            raise NotImplementedError("Method not implemented yet")

        return Z1, Z2

    def proj_fullrank(
        self, Z1: np.ndarray, Z2: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        This function provides a projection matrix U that can be applied to y, X, Z1 and Z2 to ensure its covariance
        matrix to be full-ranked. Uses variables defined in 'self', Z1 and Z2 computed in 'extract_Z2' method.
        :param Z1: numpy.ndarray
            Array of size (nz1 x p) of control simulation
        :param Z2: numpy.ndarray
            Array of size (nz1 x p) of control simulation
        :return:
        yc: numpy.ndarray
            y projected in U
        Xc: numpy.ndarray
            X projected in U
        Z1c: numpy.ndarray
            Z1 projected in U
        Z2c: numpy.ndarray
            Z2 projected in U
        """
        # M: the matrix corresponding to the temporal centering
        M = np.eye(self.nt, self.nt) - np.ones((self.nt, self.nt)) / self.nt

        # Eigen-vectors/-values of M; note that rk(M)=nt-1, so M has one eigenvalue equal to 0.
        u, d = speco(M)

        # (nt-1) first eigenvectors (ie the ones corresponding to non-zero eigenvalues)
        U = u[:, : self.nt - 1].T

        # Project all input data
        yc = np.dot(U, self.y)
        Xc = np.dot(U, self.X)
        Z1c = np.dot(U, Z1)
        Z2c = np.dot(U, Z2)

        return yc, Xc, Z1c, Z2c

    def creg(
        self,
        X: np.ndarray,
        method: Literal["ledoit", "specified"] = "ledoit",
        alpha1: float = 1e-10,
        alpha2: float = 1,
    ) -> np.ndarray:
        """
        This function compute the regularised covariance matrix estimate following the equation
        'Cr = alpha1 * Ip + alpha2 * CE' where alpha1 and alpha2 are parameters Ip is the p x p identity matrix and CE
        is the sample covariance matrix
        :param X: numpy.ndarray
            A n x p sample, meaning n iid realization of a random vector of size p.
        :param method: str
            method to compute the regularized covariance matrix
            - 'ledoit' uses Ledoit and Wolf (2003) estimate (default)
            - 'specified' uses specified values of alpha1 and alpha2
        :param alpha1: float
            Specified value for alpha1 (not used if method different than 'specified')
        :param alpha2: float
            Specified value for alpha1 (not used if method different than 'specified')
        :return:
        Cr: numpy.ndarray
            Regularized covariance matrix
        """
        n, p = X.shape

        CE = np.dot(X.T, X) / n  # sample covariance
        Ip = np.eye(p, p)

        # method for the regularised covariance matrix estimate as introduced by Ledoit & Wolf (2004) more specifically
        # on pages 379-380
        if method == "ledoit":
            m = np.trace(np.dot(CE, Ip)) / p  # first estimate in L&W
            XP = CE - m * Ip
            d2 = np.trace(np.dot(XP, XP.T)) / p  # second estimate in L&W

            bt = np.zeros(n)

            for i in range(n):
                Xi = X[i, :].reshape((1, p))
                Mi = np.dot(Xi.T, Xi)
                bt[i] = np.trace(np.dot((Mi - CE), (Mi - CE).T)) / p

            bb2 = (1.0 / n**2.0) * bt.sum()
            b2 = min(bb2, d2)  # third estimate in L&W
            a2 = d2 - b2  # fourth estimate in L&W

            alpha1 = b2 * m / d2
            alpha2 = a2 / d2

        elif method != "specified":
            raise NotImplementedError("Method not implemented yet")

        Cr = alpha1 * Ip + alpha2 * CE

        return Cr


if __name__ == "__main__":
    pass
    # observations
    # y = np.loadtxt('../example/HadCRUT4_1901-2010.asc')

    # # forcing patterns
    # X1 = np.loadtxt('../example/CNRM-CM5_ANT_1901-2010.asc')
    # X2 = np.loadtxt('../example/CNRM-CM5_NAT_1901-2010.asc')

    # X = np.stack([X1, X2], axis=1)

    # # internal variability ensemble
    # Z = np.loadtxt('../example/CTLruns.asc')
    # nt = len(y)
    # nz = int(len(Z) / nt)
    # Z = Z.reshape((nz, nt)).T  # every nt timesteps vector should be a column

    # p = PreProcess(y, X, Z)
    # Z1, Z2 = p.extract_Z2(frac=0.5)
    # yc, Xc, Z1c, Z2c = p.proj_fullrank(Z1, Z2)
    # Cr = p.creg(Z1c.T, method='ledoit')



"""
This script is designed to support optimal fingerprinting attribution analysis by providing essential data processing and analysis functions.

Qianye Su
suqianye2000@gmail.com

Reference
 - ROF:
     https://github.com/pinplex/PyDnA
     https://github.com/rafaelcabreu/attribution, by Rafael Abreu
"""

import numpy as np
import scipy.stats as stats
from typing import Dict, Any, Tuple, Literal
from .preprocess import PreProcess
from .utils import chi2_test


class AttributionModel:
    """
    A class for attribution models. The OLS implementation is heavily based on Aurelien Ribes (CNRM-GAME) scilab code
    (see more in 'preprocess.py'). Also, Aurelien Ribes model proposed in 2017 is implemented following the reference:
        Ribes, Aurelien, et al. (2017) A new statistical approach to climate change detection and attribution.
        Climate Dynamics.

    :attribute X: numpy.ndarray
        Array of size nt x nf, where nf is the number of forcings, with the model output as a timeseries
    :attribute y: numpy.ndarray
        Array of size nt with observations as a timeseries

    :method ols(self, Cf, Proj, Z2, cons_test='AT99'):
        Ordinary Least Square (OLS) estimation of beta from the linear model y = beta * X + epsilon
    """

    def __init__(self, X: np.ndarray, y: np.ndarray):
        """
        :param X: numpy.ndarray
            Array of size nt x nf, where nf is the number of forcings, with the model output as a timeseries
        :param y: numpy.ndarray
            Array of size nt with observations as a timeseries
        """
        self.y = y
        self.X = X
        self.nt = y.shape[0]  # Number of time steps
        # 1 stands for the number of spatial patterns (dealing only with timeseries)
        self.n_reduced = self.nt - 1  # Number of reduced dimensions
        self.num_forcings = X.shape[1]  # Number of forcing factors

    def ols(
        self,
        Cf: np.ndarray,
        Proj: np.ndarray,
        Z2: np.ndarray,
        cons_test: Literal["AT99"] = "AT99",
    ) -> Dict[str, np.ndarray]:
        """
        Ordinary Least Square (OLS) estimation of beta from the linear model y = beta * X + epsilon as discussed in the
        following reference:
            Allen, Myles R., and Simon FB Tett (1999) Checking for model consistency in optimal fingerprinting.
            Climate Dynamics.
        :param Cf: numpy.ndarray
            Covariance matrix. Be sure that Cf is invertible to use this model (look at PreProcess class)
        :param Proj: numpy.ndarray
            Array of zeros and ones, indicating which forcings in each simulation
        :param Z2: numpy.ndarray
            Array of size (nz1 x p) of control simulation used to compute consistency test
        :param cons_test: str
            Which consistency test to be used
            - 'AT99' the formula provided by Allen & Tett (1999) (default)
        :return:
        Beta_hat: dict
            Dictionary with estimation of beta_hat and the upper and lower confidence intervals
        """

        # computes the covariance inverse
        Cf1 = np.linalg.inv(Cf)

        _Ft = np.linalg.multi_dot([self.X.T, Cf1, self.X])
        _Ft1 = np.linalg.inv(_Ft)
        Ft = np.linalg.multi_dot([_Ft1, self.X.T, Cf1]).T

        _y = self.y.reshape((self.nt, 1))
        beta_hat = np.linalg.multi_dot([_y.T, Ft, Proj.T])

        # 1-D confidence interval
        nz2 = Z2.shape[1]
        Z2t = Z2.T
        Var_valid = np.dot(Z2t.T, Z2t) / nz2
        Var_beta_hat = np.linalg.multi_dot([Proj, Ft.T, Var_valid, Ft, Proj.T])

        beta_hat_inf = beta_hat - 2.0 * stats.t.cdf(0.95, df=nz2) * np.sqrt(
            np.diag(Var_beta_hat).T
        )
        beta_hat_sup = beta_hat + 2.0 * stats.t.cdf(0.95, df=nz2) * np.sqrt(
            np.diag(Var_beta_hat).T
        )

        # consistency check
        epsilon = _y - np.linalg.multi_dot([self.X, np.linalg.inv(Proj), beta_hat.T])

        if cons_test == "AT99":  # formula provided by Allen & Tett (1999)
            d_cons = np.linalg.multi_dot(
                [epsilon.T, np.linalg.pinv(Var_valid), epsilon]
            ) / (self.n_reduced - self.num_forcings)
            rien = stats.f.cdf(d_cons, dfn=self.n_reduced - self.num_forcings, dfd=nz2)
            pv_cons = 1 - rien

        print("Consistency test: %s p-value: %.5f" % (cons_test, pv_cons))

        Beta_hat = {
            "beta_hat": beta_hat[0],
            "beta_hat_inf": beta_hat_inf[0],
            "beta_hat_sup": beta_hat_sup[0],
        }

        return Beta_hat

    def ribes(
        self, Cxi: np.ndarray, Cy: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Aurelien Ribes model proposed in 2017 is implemented following the reference:
        Ribes, Aurelien, et al. (2017) A new statistical approach to climate change detection and attribution.
        Climate Dynamics. It considers the following set of equations:

            Y_star = sum(X_star_i) for i from 1 to nf where nf is the number of forcings
            Y = Y_star + epsilon_y
            Xi = X_star_i + epsilon_xi

        Where epislon_y ~ N(0, Cy) and epislon_xi ~ N(0, Cxi)

        :param Cxi: numpy.ndarray
            Covariance matrix for each of the forcings Xi. Should be a 3D array (nt, nt, nf)
        :param Cy: numpy.ndarray
            Covariance matrix for the observations.
        :return:
        """
        X = self.X.sum(axis=1)
        Cx = Cxi.sum(axis=0)

        # Estimate the true state of variables (y) and (Xi) y_star and X_star_i using the MLE y_star_hat and
        # Xi_star_hat, respectively
        Xi_star_hat = np.zeros(self.X.shape)
        y_star_hat = self.y + np.linalg.multi_dot(
            [Cy, np.linalg.inv(Cy + Cx), (X - self.y)]
        )
        for i in range(Xi_star_hat.shape[1]):
            Xi_star_hat[:, i] = self.X[:, i] + np.linalg.multi_dot(
                [Cxi[i], np.linalg.inv(Cy + Cx), (self.y - X)]
            )

        # calculates variance for Y_star_hat
        Cy_star_hat = np.linalg.inv(np.linalg.inv(Cy) + np.linalg.inv(Cx))

        # calculates variance for Xi_star_hat
        Cxi_star_hat = np.zeros(Cxi.shape)
        for i in range(Cxi_star_hat.shape[0]):
            Cxi_temp = Cxi * 1.0
            # sum for every j different than i
            Cxi_temp[i] = 0.0
            Cxi_sum = Cxi_temp.sum(axis=0)

            Cxi_star_hat[i] = np.linalg.inv(
                np.linalg.inv(Cxi[i]) + np.linalg.inv(Cy + Cxi_sum)
            )

        # hypothesis test: compare with chi-square distribution
        print("#" * 60)
        print(
            "Hypothesis testing p-value for Chi-2 distribution and Maximum Likelihood ..."
        )

        # (internal variability only)
        d_cons = np.linalg.multi_dot([self.y.T, np.linalg.inv(Cy), self.y])
        print(
            "%30s: %.7f (%.7f)"
            % (
                "Internal variability only",
                chi2_test(d_cons, self.nt),
                np.exp(d_cons / -2.0),
            )
        )

        # (all forcings)
        d_cons = np.linalg.multi_dot(
            [(self.y - X).T, np.linalg.inv(Cy + Cx), (self.y - X)]
        )
        print(
            "%30s: %.7f (%.7f)"
            % ("All forcings", chi2_test(d_cons, self.nt), np.exp(d_cons / -2.0))
        )

        # (individual forcings)
        for i in range(self.X.shape[1]):
            d_cons = np.linalg.multi_dot(
                [
                    (self.y - self.X[:, i]).T,
                    np.linalg.inv(Cy + Cxi[i]),
                    (self.y - self.X[:, i]),
                ]
            )
            print(
                "%30s: %.7f (%.7f)"
                % (
                    "Forcing no %d only" % (i + 1),
                    chi2_test(d_cons, self.nt),
                    np.exp(d_cons / -2.0),
                )
            )

        return y_star_hat, Xi_star_hat, Cy_star_hat, Cxi_star_hat


if __name__ == "__main__":
    pass



"""
Test skyborn.conversion module GRIB to NetCDF conversion functionality

This file contains basic tests for the conversion module.
"""

import pytest
import tempfile
import shutil
from pathlib import Path
from unittest.mock import patch, MagicMock

from skyborn.conversion import (
    convert_grib_to_nc,
    convert_grib_to_nc_simple,
    batch_convert_grib_to_nc,
    _check_grib_to_netcdf_available,
    _validate_grib_files,
    _build_grib_to_netcdf_command,
    GribToNetCDFError,
)


class TestGribToNetCDF:
    """Test GRIB to NetCDF conversion functionality"""

    def test_check_grib_to_netcdf_available(self):
        """Test checking if grib_to_netcdf is available"""
        # This test depends on whether eccodes is installed on the system
        result = _check_grib_to_netcdf_available()
        assert isinstance(result, bool)

    def test_validate_grib_files_success(self):
        """Test validating existing GRIB files"""
        with tempfile.NamedTemporaryFile(suffix=".grib", delete=False) as tmp:
            tmp_path = Path(tmp.name)

        try:
            result = _validate_grib_files([tmp_path])
            assert len(result) == 1
            assert isinstance(result[0], Path)
        finally:
            tmp_path.unlink()

    def test_validate_grib_files_not_found(self):
        """Test validating non-existent GRIB files"""
        with pytest.raises(FileNotFoundError):
            _validate_grib_files(["nonexistent.grib"])

    def test_build_grib_to_netcdf_command_basic(self):
        """Test building basic command"""
        cmd = _build_grib_to_netcdf_command(
            output_file="output.nc", grib_files=["input.grib"]
        )

        expected_parts = [
            "grib_to_netcdf",
            "-D",
            "NC_SHORT",
            "-k",
            "2",
            "-o",
            "output.nc",
            "input.grib",
        ]

        for part in expected_parts:
            assert part in cmd

    def test_build_grib_to_netcdf_command_advanced(self):
        """Test building advanced command"""
        cmd = _build_grib_to_netcdf_command(
            output_file="output.nc",
            grib_files=["input1.grib", "input2.grib"],
            ignore_keys=["type", "step"],
            split_keys=["param"],
            data_type="NC_FLOAT",
            file_kind=4,
            deflate_level=6,
            shuffle=True,
            unlimited_dimension="time",
            force=True,
        )

        expected_parts = [
            "grib_to_netcdf",
            "-I",
            "type,step",
            "-S",
            "param",
            "-D",
            "NC_FLOAT",
            "-f",
            "-k",
            "4",
            "-d",
            "6",
            "-s",
            "-u",
            "time",
            "-o",
            "output.nc",
            "input1.grib",
            "input2.grib",
        ]

        for part in expected_parts:
            assert part in cmd

    def test_build_grib_to_netcdf_command_invalid_deflate(self):
        """Test invalid compression level"""
        with pytest.raises(ValueError, match="Deflate level must be between 0 and 9"):
            _build_grib_to_netcdf_command(
                output_file="output.nc", grib_files=["input.grib"], deflate_level=10
            )

    @patch("skyborn.conversion._check_grib_to_netcdf_available")
    def test_convert_grib_to_nc_tool_not_available(self, mock_check):
        """Test when grib_to_netcdf tool is not available"""
        mock_check.return_value = False

        with pytest.raises(GribToNetCDFError, match="grib_to_netcdf command not found"):
            convert_grib_to_nc("input.grib", "output.nc")

    @patch("skyborn.conversion.subprocess.run")
    @patch("skyborn.conversion._check_grib_to_netcdf_available")
    @patch("skyborn.conversion._validate_grib_files")
    def test_convert_grib_to_nc_success(self, mock_validate, mock_check, mock_run):
        """Test successful conversion"""
        # Setup mocks
        mock_check.return_value = True
        mock_validate.return_value = [Path("input.grib")]
        mock_run.return_value = MagicMock(
            returncode=0, stdout="Conversion successful", stderr=""
        )

        result = convert_grib_to_nc("input.grib", "output.nc", verbose=False)

        assert isinstance(result, Path)
        assert result.name == "output.nc"
        mock_run.assert_called_once()

    @patch("skyborn.conversion.subprocess.run")
    @patch("skyborn.conversion._check_grib_to_netcdf_available")
    @patch("skyborn.conversion._validate_grib_files")
    def test_convert_grib_to_nc_failure(self, mock_validate, mock_check, mock_run):
        """Test conversion failure"""
        # Setup mocks
        mock_check.return_value = True
        mock_validate.return_value = [Path("input.grib")]

        # Mock command failure
        from subprocess import CalledProcessError

        mock_run.side_effect = CalledProcessError(
            returncode=1, cmd=["grib_to_netcdf"], stderr="Error message"
        )

        with pytest.raises(GribToNetCDFError, match="grib_to_netcdf failed"):
            convert_grib_to_nc("input.grib", "output.nc", verbose=False)

    @patch("skyborn.conversion.convert_grib_to_nc")
    def test_convert_grib_to_nc_simple(self, mock_convert):
        """Test simplified interface"""
        mock_convert.return_value = Path("output.nc")

        result = convert_grib_to_nc_simple(
            "input.grib", "output.nc", high_precision=True, compress=True
        )

        # Verify call arguments
        mock_convert.assert_called_once()
        call_args = mock_convert.call_args

        assert call_args[1]["data_type"] == "NC_FLOAT"  # high_precision=True
        assert call_args[1]["file_kind"] == 4  # compress=True
        assert call_args[1]["deflate_level"] == 6  # compress=True
        assert call_args[1]["shuffle"] is True  # compress=True

    def test_batch_convert_no_files(self):
        """Test batch conversion when no files are found"""
        with tempfile.TemporaryDirectory() as tmpdir:
            input_dir = Path(tmpdir) / "input"
            output_dir = Path(tmpdir) / "output"
            input_dir.mkdir()

            with pytest.raises(FileNotFoundError, match="No GRIB files found"):
                batch_convert_grib_to_nc(input_dir, output_dir)


def test_module_imports():
    """Test module imports"""
    import skyborn

    # Test direct import from skyborn
    assert hasattr(skyborn, "convert_grib_to_nc")
    assert hasattr(skyborn, "convert_grib_to_nc_simple")
    assert hasattr(skyborn, "batch_convert_grib_to_nc")
    assert hasattr(skyborn, "grib2nc")
    assert hasattr(skyborn, "grib_to_netcdf")

    # Test import from skyborn.conversion
    from skyborn.conversion import convert_grib_to_nc

    assert callable(convert_grib_to_nc)


if __name__ == "__main__":
    # Run simple tests
    print("Running skyborn.conversion module tests...")

    # Test module imports
    test_module_imports()
    print("✅ Module import tests passed")

    # Test tool checking
    available = _check_grib_to_netcdf_available()
    print(f"grib_to_netcdf tool availability: {available}")

    if available:
        print("✅ grib_to_netcdf tool is available, actual conversion is possible")
    else:
        print("⚠️  grib_to_netcdf tool is not available, please install eccodes")

    print("Tests completed!")



import numpy as np
from typing import Union, Optional

EARTH_RADIUS = 6371e3  # Earth's mean radius (m)


def calculate_gradient(
    field: np.ndarray,
    coordinates: np.ndarray,
    axis: int = -1,
    radius: float = 6371000.0,
) -> np.ndarray:
    """Calculate gradient of an arbitrary dimensional array along specified coordinates

    Args:
        field: Data field for gradient calculation, can be any dimensional array, e.g., (time, level, lat, lon)
        coordinates: Coordinate array along which to calculate the gradient, such as latitude or longitude values
        axis: Specifies the dimensional axis for gradient calculation, defaults to the last dimension (-1)
        radius: Earth radius, default is 6371000.0 meters, used for latitude gradient calculation

    Returns:
        Gradient field with the same shape as the input field
    """
    # Check if input data dimensions match
    if coordinates.size != field.shape[axis]:
        raise ValueError(
            f"Coordinate array size ({coordinates.size}) does not match field size ({field.shape[axis]}) on the specified axis"
        )

    # Determine if coordinates are latitude or longitude
    is_latitude = False
    if np.min(coordinates) >= -90 and np.max(coordinates) <= 90:
        is_latitude = True
        # For latitude, calculate actual distance (in meters)
        if is_latitude:
            # Convert latitude to actual distance
            distances = coordinates * np.pi / 180.0 * radius
        else:
            # For longitude, we need to consider the latitude effect, but this requires additional latitude information
            # Here we simply use longitude differences directly
            distances = coordinates

    # Create output array with the same shape as input
    gradient = np.zeros_like(field, dtype=float)

    # To use numpy's advanced indexing, we need to create index arrays
    ndim = field.ndim
    idx_ranges = [slice(None)] * ndim

    # Use central difference for interior points
    inner_range = slice(1, field.shape[axis] - 1)
    idx_forward = idx_ranges.copy()
    idx_forward[axis] = slice(2, field.shape[axis])

    idx_center = idx_ranges.copy()
    idx_center[axis] = inner_range

    idx_backward = idx_ranges.copy()
    idx_backward[axis] = slice(0, field.shape[axis] - 2)

    # Use vectorized operations to calculate gradient for interior points
    forward_dists = np.diff(distances[1:])
    backward_dists = np.diff(distances[:-1])
    total_dists = distances[2:] - distances[:-2]

    # Create coefficient arrays with shape suitable for broadcasting
    shape = [1] * ndim
    shape[axis] = len(forward_dists)

    a0 = forward_dists.reshape(shape)
    b0 = backward_dists.reshape(shape)
    c0 = total_dists.reshape(shape)

    # Calculate gradient using weighted difference formula
    gradient[tuple(idx_center)] = (
        b0 / a0 / c0 * field[tuple(idx_forward)]
        - a0 / b0 / c0 * field[tuple(idx_backward)]
        + (a0 - b0) / a0 / b0 * field[tuple(idx_center)]
    )

    # Handle boundary points (forward and backward differences)
    # Left boundary
    left_idx = idx_ranges.copy()
    left_idx[axis] = 0
    left_idx_plus = idx_ranges.copy()
    left_idx_plus[axis] = 1
    gradient[tuple(left_idx)] = (
        field[tuple(left_idx_plus)] - field[tuple(left_idx)]
    ) / (distances[1] - distances[0])

    # Right boundary
    right_idx = idx_ranges.copy()
    right_idx[axis] = -1
    right_idx_minus = idx_ranges.copy()
    right_idx_minus[axis] = -2
    gradient[tuple(right_idx)] = (
        field[tuple(right_idx)] - field[tuple(right_idx_minus)]
    ) / (distances[-1] - distances[-2])

    return gradient


def calculate_meridional_gradient(
    field: np.ndarray,
    latitudes: np.ndarray,
    lat_axis: int = -1,
    radius: float = 6371000.0,
) -> np.ndarray:
    """Calculate meridional gradient (gradient along latitude direction)

    Args:
        field: Data field for gradient calculation, can be any dimensional array
        latitudes: Latitude array (degrees)
        lat_axis: Specifies the axis for latitude, defaults to the last dimension (-1)
        radius: Earth radius, default is 6371000.0 meters

    Returns:
        Meridional gradient field
    """
    return calculate_gradient(field, latitudes, axis=lat_axis, radius=radius)


def calculate_vertical_gradient(
    field: np.ndarray, pressure: np.ndarray, pressure_axis: int = -3
) -> np.ndarray:
    """Calculate vertical gradient (gradient along pressure direction)

    Args:
        field: Data field for gradient calculation
        pressure: Pressure array (Pa), must be monotonically decreasing
        pressure_axis: Specifies the axis for pressure, defaults to the third-to-last dimension (-3)

    Returns:
        Vertical gradient field
    """
    return calculate_gradient(field, pressure, axis=pressure_axis, radius=None)


def calculate_zonal_gradient(
    field: np.ndarray,
    longitudes: np.ndarray,
    latitudes: np.ndarray,
    lon_axis: int = -1,
    lat_axis: int = -2,
    radius: float = 6371000.0,
) -> np.ndarray:
    """Calculate zonal gradient (gradient along longitude direction)

    Args:
        field: Data field for gradient calculation, can be any dimensional array
        longitudes: Longitude array (degrees)
        latitudes: Latitude array (degrees), used to calculate actual distance between longitudes at different latitudes
        lon_axis: Specifies the axis for longitude, defaults to the last dimension (-1)
        lat_axis: Specifies the axis for latitude, defaults to the second-to-last dimension (-2)
        radius: Earth radius, default is 6371000.0 meters

    Returns:
        Zonal gradient field
    """
    # Get latitude factor to adjust actual distance between longitudes at different latitudes
    cos_lat = np.cos(np.radians(latitudes))

    # If field is 4D (time, level, lat, lon)
    if field.ndim == 4 and lon_axis == -1 and lat_axis == -2:
        # Create a latitude factor array with shape suitable for broadcasting
        cos_lat_expanded = cos_lat.reshape(1, 1, -1, 1)

        # Convert longitudes to actual distances considering latitude
        effective_distances = np.radians(longitudes) * radius * cos_lat_expanded

        # Calculate gradient
        return calculate_gradient(field, effective_distances, axis=lon_axis, radius=1.0)

    # If field is 3D (time, lat, lon)
    elif field.ndim == 3 and lon_axis == -1 and lat_axis == -2:
        cos_lat_expanded = cos_lat.reshape(1, -1, 1)
        effective_distances = np.radians(longitudes) * radius * cos_lat_expanded
        return calculate_gradient(field, effective_distances, axis=lon_axis, radius=1.0)

    else:
        # For other dimension combinations, create appropriate broadcasting shape
        broadcast_shape = [1] * field.ndim
        broadcast_shape[lat_axis] = len(latitudes)
        cos_lat_expanded = cos_lat.reshape(broadcast_shape)

        # Create effective distance array
        effective_longitudes = np.radians(longitudes) * radius

        # Calculate gradient for each latitude
        result = np.zeros_like(field)

        # Loop through each latitude (implementation depends on specific data structure, may need adjustment)
        for i in range(len(latitudes)):
            idx = [slice(None)] * field.ndim
            idx[lat_axis] = i

            # Adjust longitude distance for current latitude
            current_effective_dist = effective_longitudes * cos_lat[i]

            # Calculate gradient for current latitude
            result[tuple(idx)] = calculate_gradient(
                field[tuple(idx)], current_effective_dist, axis=lon_axis, radius=1.0
            )

        return result



"""
Data format conversion utilities for the skyborn library.

This module provides functions for converting between different atmospheric
data formats, including GRIB to NetCDF conversion using eccodes.

Author: Qianye Su
Email: suqianye2000@gmail.com
"""

from .grib_to_netcdf import (
    convert_grib_to_nc,
    convert_grib_to_nc_simple,
    batch_convert_grib_to_nc,
    grib2nc,
    grib_to_netcdf,
    GribToNetCDFError,
)

__all__ = [
    "convert_grib_to_nc",
    "convert_grib_to_nc_simple",
    "batch_convert_grib_to_nc",
    "grib2nc",
    "grib_to_netcdf",
    "GribToNetCDFError",
]
